#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import tempfile
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from rapidtide.workflows.ccorrica import _get_parser, ccorrica

# ---- helpers ----


def _make_args(**overrides):
    """Build a minimal args Namespace for ccorrica with sensible defaults."""
    defaults = dict(
        timecoursefile="input.txt",
        outputroot="/tmp/test_ccorrica_out",
        samplerate=2.0,
        oversampfactor=1,
        detrendorder=3,
        windowfunc="hamming",
        corrweighting="phat",
        debug=False,
        # filter opts expected by postprocessfilteropts
        filterband="lfo",
        passvec=None,
        stopvec=None,
    )
    defaults.update(overrides)
    return argparse.Namespace(**defaults)


def _make_mock_prefilter():
    """Create a mock prefilter whose apply() returns input unchanged."""
    prefilter = MagicMock()
    prefilter.apply.side_effect = lambda fs, data: data
    return prefilter


def _make_tc_data(numcomponents=3, tclen=100):
    """Create synthetic timecourse data (numcomponents x tclen)."""
    rng = np.random.RandomState(42)
    return rng.randn(numcomponents, tclen).astype(float)


def _capture_writenpvecs():
    """Capture writenpvecs calls, copying data at call time."""
    captured = []

    def _side_effect(data, filename, **kwargs):
        captured.append((np.array(data, copy=True), filename))

    return _side_effect, captured


def _capture_savetonifti():
    """Capture savetonifti calls, copying data at call time."""
    captured = []

    def _side_effect(data, hdr, filename, **kwargs):
        captured.append((np.array(data, copy=True), filename))

    return _side_effect, captured


def _run_ccorrica(args, tcdata=None, file_samplerate=None, numcomponents=3, tclen=100):
    """Run ccorrica with full mocking of external dependencies.

    Parameters
    ----------
    args : Namespace
        Arguments to pass to ccorrica.
    tcdata : ndarray, optional
        Timecourse data. If None, synthetic data is generated.
    file_samplerate : float or None
        Sample rate returned by readvectorsfromtextfile.  None means
        the file doesn't specify a sample rate.
    numcomponents : int
        Number of components (used when tcdata is None).
    tclen : int
        Length of timecourses (used when tcdata is None).

    Returns
    -------
    dict with keys: 'writenpvecs', 'savetonifti', 'args_out'
    """
    if tcdata is None:
        tcdata = _make_tc_data(numcomponents, tclen)
    else:
        numcomponents, tclen = tcdata.shape

    prefilter = _make_mock_prefilter()

    wnp_effect, wnp_captured = _capture_writenpvecs()
    stn_effect, stn_captured = _capture_savetonifti()

    # findmaxlag_gauss returns 8-tuple
    def mock_findmaxlag_gauss(x, y, lagmin, lagmax, widthmax, **kwargs):
        midx = len(y) // 2
        return (midx, 0.0, 0.5, 1.0, 1.0, 0, 0, len(y))

    # symmetrize: return input unchanged
    def mock_symmetrize(matrix, zerodiagonal=False, antisymmetric=False):
        result = np.array(matrix, copy=True)
        if zerodiagonal:
            np.fill_diagonal(result, 0.0)
        return result

    # fastcorrelate: return a properly-sized cross-correlation array
    xcorrlen = 2 * tclen - 1

    def mock_fastcorrelate(a, b, usefft=True, weighting=None, zeropadding=0, displayplots=False):
        rng = np.random.RandomState(hash((a.tobytes()[:16], b.tobytes()[:16])) % (2**31))
        result = rng.randn(xcorrlen) * 0.1
        result[xcorrlen // 2] = 0.5  # peak at zero lag
        return result

    # pearsonr mock
    mock_pearsonr_result = MagicMock()
    mock_pearsonr_result.statistic = 0.3

    # nib.Nifti1Image mock — returns an object whose .header supports item access
    def mock_nifti1image(data, affine):
        mock_img = MagicMock()
        mock_hdr = {}
        mock_hdr["pixdim"] = np.zeros(8)
        mock_img.header = mock_hdr
        return mock_img

    with (
        patch(
            "rapidtide.workflows.ccorrica.pf.postprocessfilteropts",
            return_value=(args, prefilter),
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_io.readvectorsfromtextfile",
            return_value=(file_samplerate, 0.0, None, tcdata, False, "text"),
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_resample.upsample",
            side_effect=lambda data, fs_in, fs_out, intfac=False: np.repeat(
                data, int(fs_out / fs_in)
            ),
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_math.stdnormalize",
            side_effect=lambda x: x / (np.std(x) if np.std(x) > 0 else 1.0),
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_math.corrnormalize",
            side_effect=lambda x, detrendorder=0, windowfunc="hamming": x
            / (np.std(x) if np.std(x) > 0 else 1.0),
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_corr.fastcorrelate",
            side_effect=mock_fastcorrelate,
        ),
        patch(
            "rapidtide.workflows.ccorrica.pearsonr",
            return_value=mock_pearsonr_result,
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_fit.findmaxlag_gauss",
            side_effect=mock_findmaxlag_gauss,
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_stats.symmetrize",
            side_effect=mock_symmetrize,
        ),
        patch(
            "rapidtide.workflows.ccorrica.nib.Nifti1Image",
            side_effect=mock_nifti1image,
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_io.savetonifti",
            side_effect=stn_effect,
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_io.writenpvecs",
            side_effect=wnp_effect,
        ),
    ):
        ccorrica(args)

    return {
        "writenpvecs": wnp_captured,
        "savetonifti": stn_captured,
    }


# ---- _get_parser tests ----


def test_parser_required_args(debug=False):
    """Parser should accept the two required positional args."""
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        parser = _get_parser()
        args = parser.parse_args([f.name, "outroot"])
        assert args.timecoursefile == f.name
        assert args.outputroot == "outroot"
        if debug:
            print("test_parser_required_args passed")


def test_parser_defaults(debug=False):
    """Parser defaults should match expected values."""
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        parser = _get_parser()
        args = parser.parse_args([f.name, "outroot"])
        assert args.samplerate == "auto"
        assert args.oversampfactor == 1
        assert args.detrendorder == 3
        assert args.corrweighting == "phat"
        assert args.debug is False
        if debug:
            print("test_parser_defaults passed")


def test_parser_samplerate(debug=False):
    """Parser should accept --samplerate."""
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        parser = _get_parser()
        args = parser.parse_args([f.name, "outroot", "--samplerate", "4.0"])
        assert args.samplerate == 4.0
        if debug:
            print("test_parser_samplerate passed")


def test_parser_sampletstep(debug=False):
    """Parser should accept --sampletstep and invert it to samplerate."""
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        parser = _get_parser()
        args = parser.parse_args([f.name, "outroot", "--sampletstep", "0.5"])
        assert abs(args.samplerate - 2.0) < 1e-6
        if debug:
            print("test_parser_sampletstep passed")


def test_parser_mutually_exclusive_rate(debug=False):
    """Parser should reject both --samplerate and --sampletstep together."""
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        parser = _get_parser()
        with pytest.raises(SystemExit):
            parser.parse_args([f.name, "outroot", "--samplerate", "2.0", "--sampletstep", "0.5"])
        if debug:
            print("test_parser_mutually_exclusive_rate passed")


def test_parser_corrweighting_choices(debug=False):
    """Parser should accept valid corrweighting choices and reject invalid ones."""
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        parser = _get_parser()
        for choice in ["None", "phat", "liang", "eckart"]:
            args = parser.parse_args([f.name, "outroot", "--corrweighting", choice])
            assert args.corrweighting == choice
        with pytest.raises(SystemExit):
            parser.parse_args([f.name, "outroot", "--corrweighting", "invalid"])
        if debug:
            print("test_parser_corrweighting_choices passed")


def test_parser_detrendorder(debug=False):
    """Parser should accept --detrendorder."""
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        parser = _get_parser()
        args = parser.parse_args([f.name, "outroot", "--detrendorder", "5"])
        assert args.detrendorder == 5
        if debug:
            print("test_parser_detrendorder passed")


def test_parser_oversampfactor(debug=False):
    """Parser should accept --oversampfactor including negative values."""
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        parser = _get_parser()
        args = parser.parse_args([f.name, "outroot", "--oversampfactor", "-1"])
        assert args.oversampfactor == -1
        if debug:
            print("test_parser_oversampfactor passed")


def test_parser_debug_flag(debug=False):
    """Parser should accept --debug flag."""
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        parser = _get_parser()
        args = parser.parse_args([f.name, "outroot", "--debug"])
        assert args.debug is True
        if debug:
            print("test_parser_debug_flag passed")


# ---- ccorrica happy path tests ----


def test_ccorrica_basic_run(debug=False):
    """ccorrica should run without errors with basic inputs."""
    args = _make_args()
    result = _run_ccorrica(args, numcomponents=3, tclen=100)
    assert len(result["writenpvecs"]) > 0
    assert len(result["savetonifti"]) > 0
    if debug:
        print("test_ccorrica_basic_run passed")


def test_ccorrica_output_files(debug=False):
    """ccorrica should write all expected output files."""
    args = _make_args(outputroot="/tmp/cctest")
    result = _run_ccorrica(args, numcomponents=2, tclen=50)

    # Check savetonifti output files
    stn_files = [entry[1] for entry in result["savetonifti"]]
    expected_nifti_suffixes = [
        "_xcorr",
        "_pxcorr",
        "_corrmax",
        "_corrlag",
        "_corrwidth",
        "_corrmask",
    ]
    for suffix in expected_nifti_suffixes:
        assert any(suffix in f for f in stn_files), f"Missing nifti file with suffix {suffix}"

    # Check writenpvecs output files
    wnp_files = [entry[1] for entry in result["writenpvecs"]]
    expected_txt_suffixes = [
        "_filtereddata.txt",
        "_corrmax.txt",
        "_corrlag.txt",
        "_corrwidth.txt",
        "_corrmask.txt",
        "_reformdata.txt",
    ]
    for suffix in expected_txt_suffixes:
        assert any(suffix in f for f in wnp_files), f"Missing text file with suffix {suffix}"

    if debug:
        print("test_ccorrica_output_files passed")


def test_ccorrica_output_shapes(debug=False):
    """Output data should have correct shapes for given numcomponents."""
    numcomponents = 3
    tclen = 60
    args = _make_args()
    result = _run_ccorrica(args, numcomponents=numcomponents, tclen=tclen)

    # Find corrmax output in savetonifti calls
    for data, fname in result["savetonifti"]:
        if "_corrmax" in fname and "_pxcorr" not in fname:
            assert data.shape[0] == numcomponents
            assert data.shape[1] == numcomponents
        if "_pxcorr" in fname:
            assert data.shape[0] == numcomponents
            assert data.shape[1] == numcomponents
            assert data.shape[3] == tclen

    # Find corrmax txt output
    for data, fname in result["writenpvecs"]:
        if "_corrmax.txt" in fname:
            assert data.shape == (numcomponents, numcomponents)

    if debug:
        print("test_ccorrica_output_shapes passed")


def test_ccorrica_filtereddata_shape(debug=False):
    """Filtered data output should have shape (numcomponents, tclen)."""
    numcomponents = 4
    tclen = 80
    args = _make_args()
    result = _run_ccorrica(args, numcomponents=numcomponents, tclen=tclen)

    for data, fname in result["writenpvecs"]:
        if "_filtereddata.txt" in fname:
            assert data.shape == (numcomponents, tclen)
            break
    else:
        pytest.fail("_filtereddata.txt not found in output")

    if debug:
        print("test_ccorrica_filtereddata_shape passed")


def test_ccorrica_reformdata_shape(debug=False):
    """Reform data output should have shape (numcomponents, tclen)."""
    numcomponents = 3
    tclen = 50
    args = _make_args()
    result = _run_ccorrica(args, numcomponents=numcomponents, tclen=tclen)

    for data, fname in result["writenpvecs"]:
        if "_reformdata.txt" in fname:
            assert data.shape == (numcomponents, tclen)
            break
    else:
        pytest.fail("_reformdata.txt not found in output")

    if debug:
        print("test_ccorrica_reformdata_shape passed")


def test_ccorrica_single_component(debug=False):
    """ccorrica should handle single component data."""
    args = _make_args()
    result = _run_ccorrica(args, numcomponents=1, tclen=50)

    # Should still produce output files
    assert len(result["writetonifti"] if "writetonifti" in result else result["savetonifti"]) > 0
    assert len(result["writenpvecs"]) > 0

    # corrmax should be 1x1
    for data, fname in result["writenpvecs"]:
        if "_corrmax.txt" in fname:
            assert data.shape == (1, 1)

    if debug:
        print("test_ccorrica_single_component passed")


def test_ccorrica_two_components(debug=False):
    """ccorrica should correctly handle 2 components."""
    args = _make_args()
    result = _run_ccorrica(args, numcomponents=2, tclen=40)

    for data, fname in result["writenpvecs"]:
        if "_corrlag.txt" in fname:
            assert data.shape == (2, 2)

    if debug:
        print("test_ccorrica_two_components passed")


def test_ccorrica_xcorr_nifti_shape(debug=False):
    """Cross-correlation nifti should have correct windowed length."""
    numcomponents = 2
    tclen = 50
    args = _make_args(samplerate=2.0)
    result = _run_ccorrica(args, numcomponents=numcomponents, tclen=tclen)

    for data, fname in result["savetonifti"]:
        if "_xcorr" in fname and "_pxcorr" not in fname:
            # Shape should be (numcomponents, numcomponents, 1, corrwin)
            assert data.shape[0] == numcomponents
            assert data.shape[1] == numcomponents
            assert data.shape[2] == 1
            # The window length is searchend - searchstart
            Fs = 2.0
            searchrange = 15.0
            halfwindow = int(searchrange * Fs)
            xcorrlen = 2 * tclen - 1
            corrzero = xcorrlen // 2
            corrwin = 2 * halfwindow
            assert data.shape[3] == corrwin
            break
    else:
        pytest.fail("_xcorr nifti not found")

    if debug:
        print("test_ccorrica_xcorr_nifti_shape passed")


# ---- ccorrica sample rate handling ----


def test_ccorrica_samplerate_from_file(debug=False):
    """ccorrica should use sample rate from file when available."""
    args = _make_args(samplerate="auto")
    # file_samplerate=4.0 should be used
    result = _run_ccorrica(args, file_samplerate=4.0, numcomponents=2, tclen=50)
    assert len(result["savetonifti"]) > 0
    if debug:
        print("test_ccorrica_samplerate_from_file passed")


def test_ccorrica_samplerate_from_args(debug=False):
    """ccorrica should use args.samplerate when file doesn't provide one."""
    args = _make_args(samplerate=3.0)
    # file_samplerate=None means no rate in file
    result = _run_ccorrica(args, file_samplerate=None, numcomponents=2, tclen=50)
    assert len(result["savetonifti"]) > 0
    if debug:
        print("test_ccorrica_samplerate_from_args passed")


def test_ccorrica_missing_samplerate_exits(debug=False):
    """ccorrica should exit when samplerate is 'auto' and file has none."""
    args = _make_args(samplerate="auto")

    prefilter = _make_mock_prefilter()
    tcdata = _make_tc_data(2, 50)

    with (
        patch(
            "rapidtide.workflows.ccorrica.pf.postprocessfilteropts",
            return_value=(args, prefilter),
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_io.readvectorsfromtextfile",
            return_value=(None, 0.0, None, tcdata, False, "text"),
        ),
        patch(
            "rapidtide.workflows.ccorrica.sys.exit",
            side_effect=SystemExit,
        ) as mock_exit,
    ):
        with pytest.raises(SystemExit):
            ccorrica(args)
        mock_exit.assert_called_once()

    if debug:
        print("test_ccorrica_missing_samplerate_exits passed")


# ---- oversampling tests ----


def test_ccorrica_oversample(debug=False):
    """ccorrica should upsample data when oversampfactor > 1."""
    args = _make_args(oversampfactor=2, samplerate=2.0)
    numcomponents = 2
    tclen = 50

    tcdata = _make_tc_data(numcomponents, tclen)

    prefilter = _make_mock_prefilter()
    wnp_effect, wnp_captured = _capture_writenpvecs()
    stn_effect, stn_captured = _capture_savetonifti()

    upsample_calls = []

    def mock_upsample(data, fs_in, fs_out, intfac=False):
        upsample_calls.append((fs_in, fs_out))
        return np.repeat(data, int(fs_out / fs_in))

    def mock_findmaxlag_gauss(x, y, lagmin, lagmax, widthmax, **kwargs):
        midx = len(y) // 2
        return (midx, 0.0, 0.5, 1.0, 1.0, 0, 0, len(y))

    def mock_symmetrize(matrix, zerodiagonal=False, antisymmetric=False):
        result = np.array(matrix, copy=True)
        if zerodiagonal:
            np.fill_diagonal(result, 0.0)
        return result

    def mock_nifti1image(data, affine):
        mock_img = MagicMock()
        mock_hdr = {}
        mock_hdr["pixdim"] = np.zeros(8)
        mock_img.header = mock_hdr
        return mock_img

    mock_pearsonr_result = MagicMock()
    mock_pearsonr_result.statistic = 0.3

    new_tclen = tclen * 2  # oversampled
    xcorrlen = 2 * new_tclen - 1

    def mock_fastcorrelate(a, b, usefft=True, weighting=None, zeropadding=0, displayplots=False):
        return np.zeros(xcorrlen)

    with (
        patch(
            "rapidtide.workflows.ccorrica.pf.postprocessfilteropts", return_value=(args, prefilter)
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_io.readvectorsfromtextfile",
            return_value=(None, 0.0, None, tcdata, False, "text"),
        ),
        patch("rapidtide.workflows.ccorrica.tide_resample.upsample", side_effect=mock_upsample),
        patch("rapidtide.workflows.ccorrica.tide_math.stdnormalize", side_effect=lambda x: x),
        patch(
            "rapidtide.workflows.ccorrica.tide_math.corrnormalize", side_effect=lambda x, **kw: x
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_corr.fastcorrelate", side_effect=mock_fastcorrelate
        ),
        patch("rapidtide.workflows.ccorrica.pearsonr", return_value=mock_pearsonr_result),
        patch(
            "rapidtide.workflows.ccorrica.tide_fit.findmaxlag_gauss",
            side_effect=mock_findmaxlag_gauss,
        ),
        patch("rapidtide.workflows.ccorrica.tide_stats.symmetrize", side_effect=mock_symmetrize),
        patch("rapidtide.workflows.ccorrica.nib.Nifti1Image", side_effect=mock_nifti1image),
        patch("rapidtide.workflows.ccorrica.tide_io.savetonifti", side_effect=stn_effect),
        patch("rapidtide.workflows.ccorrica.tide_io.writenpvecs", side_effect=wnp_effect),
    ):
        ccorrica(args)

    # upsample should have been called for each component
    assert len(upsample_calls) == numcomponents
    for fs_in, fs_out in upsample_calls:
        assert fs_in == 2.0
        assert fs_out == 4.0  # 2.0 * oversampfactor(2)

    if debug:
        print("test_ccorrica_oversample passed")


def test_ccorrica_auto_oversampfactor(debug=False):
    """ccorrica should auto-compute oversampfactor when negative."""
    # With samplerate=1.0, sampletime=1.0, oversampfactor should be max(ceil(1.0/0.5), 1) = 2
    args = _make_args(oversampfactor=-1, samplerate=1.0)
    numcomponents = 2
    tclen = 50

    tcdata = _make_tc_data(numcomponents, tclen)
    prefilter = _make_mock_prefilter()
    wnp_effect, wnp_captured = _capture_writenpvecs()
    stn_effect, stn_captured = _capture_savetonifti()

    upsample_calls = []

    def mock_upsample(data, fs_in, fs_out, intfac=False):
        upsample_calls.append((fs_in, fs_out))
        return np.repeat(data, int(fs_out / fs_in))

    def mock_findmaxlag_gauss(x, y, lagmin, lagmax, widthmax, **kwargs):
        midx = len(y) // 2
        return (midx, 0.0, 0.5, 1.0, 1.0, 0, 0, len(y))

    def mock_symmetrize(matrix, zerodiagonal=False, antisymmetric=False):
        result = np.array(matrix, copy=True)
        if zerodiagonal:
            np.fill_diagonal(result, 0.0)
        return result

    def mock_nifti1image(data, affine):
        mock_img = MagicMock()
        mock_hdr = {}
        mock_hdr["pixdim"] = np.zeros(8)
        mock_img.header = mock_hdr
        return mock_img

    mock_pearsonr_result = MagicMock()
    mock_pearsonr_result.statistic = 0.3

    # auto factor for sampletime=1.0: ceil(1.0/0.5) = 2
    new_tclen = tclen * 2
    xcorrlen = 2 * new_tclen - 1

    def mock_fastcorrelate(a, b, usefft=True, weighting=None, zeropadding=0, displayplots=False):
        return np.zeros(xcorrlen)

    with (
        patch(
            "rapidtide.workflows.ccorrica.pf.postprocessfilteropts", return_value=(args, prefilter)
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_io.readvectorsfromtextfile",
            return_value=(1.0, 0.0, None, tcdata, False, "text"),
        ),
        patch("rapidtide.workflows.ccorrica.tide_resample.upsample", side_effect=mock_upsample),
        patch("rapidtide.workflows.ccorrica.tide_math.stdnormalize", side_effect=lambda x: x),
        patch(
            "rapidtide.workflows.ccorrica.tide_math.corrnormalize", side_effect=lambda x, **kw: x
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_corr.fastcorrelate", side_effect=mock_fastcorrelate
        ),
        patch("rapidtide.workflows.ccorrica.pearsonr", return_value=mock_pearsonr_result),
        patch(
            "rapidtide.workflows.ccorrica.tide_fit.findmaxlag_gauss",
            side_effect=mock_findmaxlag_gauss,
        ),
        patch("rapidtide.workflows.ccorrica.tide_stats.symmetrize", side_effect=mock_symmetrize),
        patch("rapidtide.workflows.ccorrica.nib.Nifti1Image", side_effect=mock_nifti1image),
        patch("rapidtide.workflows.ccorrica.tide_io.savetonifti", side_effect=stn_effect),
        patch("rapidtide.workflows.ccorrica.tide_io.writenpvecs", side_effect=wnp_effect),
    ):
        ccorrica(args)

    # Should have called upsample (oversampfactor auto-set to 2)
    assert len(upsample_calls) == numcomponents
    assert args.oversampfactor == 2

    if debug:
        print("test_ccorrica_auto_oversampfactor passed")


# ---- ccorrica option variation tests ----


def test_ccorrica_corrweighting_none(debug=False):
    """ccorrica should work with corrweighting='None'."""
    args = _make_args(corrweighting="None")
    result = _run_ccorrica(args, numcomponents=2, tclen=40)
    assert len(result["savetonifti"]) > 0
    if debug:
        print("test_ccorrica_corrweighting_none passed")


def test_ccorrica_detrendorder_zero(debug=False):
    """ccorrica should work with detrendorder=0 (no detrending)."""
    args = _make_args(detrendorder=0)
    result = _run_ccorrica(args, numcomponents=2, tclen=40)
    assert len(result["savetonifti"]) > 0
    if debug:
        print("test_ccorrica_detrendorder_zero passed")


def test_ccorrica_debug_mode(debug=False):
    """ccorrica should run with debug=True."""
    args = _make_args(debug=True)
    result = _run_ccorrica(args, numcomponents=2, tclen=40)
    assert len(result["savetonifti"]) > 0
    if debug:
        print("test_ccorrica_debug_mode passed")


def test_ccorrica_window_func_hann(debug=False):
    """ccorrica should work with windowfunc='hann'."""
    args = _make_args(windowfunc="hann")
    result = _run_ccorrica(args, numcomponents=2, tclen=40)
    assert len(result["savetonifti"]) > 0
    if debug:
        print("test_ccorrica_window_func_hann passed")


# ---- symmetrization tests ----


def test_ccorrica_symmetrize_called(debug=False):
    """ccorrica should call symmetrize for corrmax, corrlag, corrwidth, corrmask."""
    args = _make_args()
    numcomponents = 3
    tclen = 50
    tcdata = _make_tc_data(numcomponents, tclen)
    prefilter = _make_mock_prefilter()

    symmetrize_calls = []

    def mock_symmetrize(matrix, zerodiagonal=False, antisymmetric=False):
        symmetrize_calls.append(
            {
                "zerodiagonal": zerodiagonal,
                "antisymmetric": antisymmetric,
            }
        )
        result = np.array(matrix, copy=True)
        if zerodiagonal:
            np.fill_diagonal(result, 0.0)
        return result

    wnp_effect, _ = _capture_writenpvecs()
    stn_effect, _ = _capture_savetonifti()

    xcorrlen = 2 * tclen - 1

    def mock_fastcorrelate(a, b, usefft=True, weighting=None, zeropadding=0, displayplots=False):
        return np.zeros(xcorrlen)

    def mock_findmaxlag_gauss(x, y, lagmin, lagmax, widthmax, **kwargs):
        midx = len(y) // 2
        return (midx, 0.0, 0.5, 1.0, 1.0, 0, 0, len(y))

    mock_pearsonr_result = MagicMock()
    mock_pearsonr_result.statistic = 0.3

    def mock_nifti1image(data, affine):
        mock_img = MagicMock()
        mock_hdr = {}
        mock_hdr["pixdim"] = np.zeros(8)
        mock_img.header = mock_hdr
        return mock_img

    with (
        patch(
            "rapidtide.workflows.ccorrica.pf.postprocessfilteropts", return_value=(args, prefilter)
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_io.readvectorsfromtextfile",
            return_value=(2.0, 0.0, None, tcdata, False, "text"),
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_resample.upsample",
            side_effect=lambda d, fi, fo, intfac=False: d,
        ),
        patch("rapidtide.workflows.ccorrica.tide_math.stdnormalize", side_effect=lambda x: x),
        patch(
            "rapidtide.workflows.ccorrica.tide_math.corrnormalize", side_effect=lambda x, **kw: x
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_corr.fastcorrelate", side_effect=mock_fastcorrelate
        ),
        patch("rapidtide.workflows.ccorrica.pearsonr", return_value=mock_pearsonr_result),
        patch(
            "rapidtide.workflows.ccorrica.tide_fit.findmaxlag_gauss",
            side_effect=mock_findmaxlag_gauss,
        ),
        patch("rapidtide.workflows.ccorrica.tide_stats.symmetrize", side_effect=mock_symmetrize),
        patch("rapidtide.workflows.ccorrica.nib.Nifti1Image", side_effect=mock_nifti1image),
        patch("rapidtide.workflows.ccorrica.tide_io.savetonifti", side_effect=stn_effect),
        patch("rapidtide.workflows.ccorrica.tide_io.writenpvecs", side_effect=wnp_effect),
    ):
        ccorrica(args)

    # symmetrize should be called 4 times:
    # corrmax (zerodiagonal=True), corrlag (antisymmetric=True),
    # corrwidth (neither), corrmask (zerodiagonal=True)
    assert len(symmetrize_calls) == 4
    assert symmetrize_calls[0]["zerodiagonal"] is True  # corrmax
    assert symmetrize_calls[1]["antisymmetric"] is True  # corrlag
    assert symmetrize_calls[2]["zerodiagonal"] is False  # corrwidth
    assert symmetrize_calls[2]["antisymmetric"] is False  # corrwidth
    assert symmetrize_calls[3]["zerodiagonal"] is True  # corrmask

    if debug:
        print("test_ccorrica_symmetrize_called passed")


# ---- correlation computation tests ----


def test_ccorrica_fastcorrelate_call_count(debug=False):
    """fastcorrelate should be called numcomponents^2 times (all pairs)."""
    numcomponents = 3
    tclen = 40
    args = _make_args()
    tcdata = _make_tc_data(numcomponents, tclen)
    prefilter = _make_mock_prefilter()

    correlate_count = [0]
    xcorrlen = 2 * tclen - 1

    def mock_fastcorrelate(a, b, usefft=True, weighting=None, zeropadding=0, displayplots=False):
        correlate_count[0] += 1
        return np.zeros(xcorrlen)

    def mock_findmaxlag_gauss(x, y, lagmin, lagmax, widthmax, **kwargs):
        midx = len(y) // 2
        return (midx, 0.0, 0.5, 1.0, 1.0, 0, 0, len(y))

    def mock_symmetrize(matrix, zerodiagonal=False, antisymmetric=False):
        return np.array(matrix, copy=True)

    mock_pearsonr_result = MagicMock()
    mock_pearsonr_result.statistic = 0.3

    wnp_effect, _ = _capture_writenpvecs()
    stn_effect, _ = _capture_savetonifti()

    def mock_nifti1image(data, affine):
        mock_img = MagicMock()
        mock_hdr = {}
        mock_hdr["pixdim"] = np.zeros(8)
        mock_img.header = mock_hdr
        return mock_img

    with (
        patch(
            "rapidtide.workflows.ccorrica.pf.postprocessfilteropts", return_value=(args, prefilter)
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_io.readvectorsfromtextfile",
            return_value=(2.0, 0.0, None, tcdata, False, "text"),
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_resample.upsample",
            side_effect=lambda d, fi, fo, intfac=False: d,
        ),
        patch("rapidtide.workflows.ccorrica.tide_math.stdnormalize", side_effect=lambda x: x),
        patch(
            "rapidtide.workflows.ccorrica.tide_math.corrnormalize", side_effect=lambda x, **kw: x
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_corr.fastcorrelate", side_effect=mock_fastcorrelate
        ),
        patch("rapidtide.workflows.ccorrica.pearsonr", return_value=mock_pearsonr_result),
        patch(
            "rapidtide.workflows.ccorrica.tide_fit.findmaxlag_gauss",
            side_effect=mock_findmaxlag_gauss,
        ),
        patch("rapidtide.workflows.ccorrica.tide_stats.symmetrize", side_effect=mock_symmetrize),
        patch("rapidtide.workflows.ccorrica.nib.Nifti1Image", side_effect=mock_nifti1image),
        patch("rapidtide.workflows.ccorrica.tide_io.savetonifti", side_effect=stn_effect),
        patch("rapidtide.workflows.ccorrica.tide_io.writenpvecs", side_effect=wnp_effect),
    ):
        ccorrica(args)

    assert correlate_count[0] == numcomponents * numcomponents

    if debug:
        print("test_ccorrica_fastcorrelate_call_count passed")


def test_ccorrica_pearsonr_call_count(debug=False):
    """pearsonr should be called numcomponents^2 times."""
    numcomponents = 2
    tclen = 40
    args = _make_args()
    result = _run_ccorrica(args, numcomponents=numcomponents, tclen=tclen)

    # Check pxcorr data has correct shape
    for data, fname in result["savetonifti"]:
        if "_pxcorr" in fname:
            assert data.shape == (numcomponents, numcomponents, 1, tclen)

    if debug:
        print("test_ccorrica_pearsonr_call_count passed")


def test_ccorrica_findmaxlag_gauss_called(debug=False):
    """findmaxlag_gauss should be called for each component pair."""
    numcomponents = 2
    tclen = 40
    args = _make_args()
    tcdata = _make_tc_data(numcomponents, tclen)
    prefilter = _make_mock_prefilter()

    gauss_calls = [0]
    xcorrlen = 2 * tclen - 1

    def mock_fastcorrelate(a, b, usefft=True, weighting=None, zeropadding=0, displayplots=False):
        return np.zeros(xcorrlen)

    def mock_findmaxlag_gauss(x, y, lagmin, lagmax, widthmax, **kwargs):
        gauss_calls[0] += 1
        midx = len(y) // 2
        return (midx, 0.0, 0.5, 1.0, 1.0, 0, 0, len(y))

    def mock_symmetrize(matrix, zerodiagonal=False, antisymmetric=False):
        return np.array(matrix, copy=True)

    mock_pearsonr_result = MagicMock()
    mock_pearsonr_result.statistic = 0.3

    wnp_effect, _ = _capture_writenpvecs()
    stn_effect, _ = _capture_savetonifti()

    def mock_nifti1image(data, affine):
        mock_img = MagicMock()
        mock_hdr = {}
        mock_hdr["pixdim"] = np.zeros(8)
        mock_img.header = mock_hdr
        return mock_img

    with (
        patch(
            "rapidtide.workflows.ccorrica.pf.postprocessfilteropts", return_value=(args, prefilter)
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_io.readvectorsfromtextfile",
            return_value=(2.0, 0.0, None, tcdata, False, "text"),
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_resample.upsample",
            side_effect=lambda d, fi, fo, intfac=False: d,
        ),
        patch("rapidtide.workflows.ccorrica.tide_math.stdnormalize", side_effect=lambda x: x),
        patch(
            "rapidtide.workflows.ccorrica.tide_math.corrnormalize", side_effect=lambda x, **kw: x
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_corr.fastcorrelate", side_effect=mock_fastcorrelate
        ),
        patch("rapidtide.workflows.ccorrica.pearsonr", return_value=mock_pearsonr_result),
        patch(
            "rapidtide.workflows.ccorrica.tide_fit.findmaxlag_gauss",
            side_effect=mock_findmaxlag_gauss,
        ),
        patch("rapidtide.workflows.ccorrica.tide_stats.symmetrize", side_effect=mock_symmetrize),
        patch("rapidtide.workflows.ccorrica.nib.Nifti1Image", side_effect=mock_nifti1image),
        patch("rapidtide.workflows.ccorrica.tide_io.savetonifti", side_effect=stn_effect),
        patch("rapidtide.workflows.ccorrica.tide_io.writenpvecs", side_effect=wnp_effect),
    ):
        ccorrica(args)

    assert gauss_calls[0] == numcomponents * numcomponents

    if debug:
        print("test_ccorrica_findmaxlag_gauss_called passed")


# ---- nifti header tests ----


def test_ccorrica_nifti_pixdim_set(debug=False):
    """Nifti headers should have pixdim[4] set to sampletime."""
    args = _make_args(samplerate=4.0)
    numcomponents = 2
    tclen = 40

    tcdata = _make_tc_data(numcomponents, tclen)
    prefilter = _make_mock_prefilter()

    nifti_headers = []

    def mock_nifti1image(data, affine):
        mock_img = MagicMock()
        mock_hdr = {}
        mock_hdr["pixdim"] = np.zeros(8)
        mock_img.header = mock_hdr
        nifti_headers.append(mock_hdr)
        return mock_img

    xcorrlen = 2 * tclen - 1

    def mock_fastcorrelate(a, b, usefft=True, weighting=None, zeropadding=0, displayplots=False):
        return np.zeros(xcorrlen)

    def mock_findmaxlag_gauss(x, y, lagmin, lagmax, widthmax, **kwargs):
        midx = len(y) // 2
        return (midx, 0.0, 0.5, 1.0, 1.0, 0, 0, len(y))

    def mock_symmetrize(matrix, zerodiagonal=False, antisymmetric=False):
        return np.array(matrix, copy=True)

    mock_pearsonr_result = MagicMock()
    mock_pearsonr_result.statistic = 0.3

    wnp_effect, _ = _capture_writenpvecs()
    stn_effect, _ = _capture_savetonifti()

    with (
        patch(
            "rapidtide.workflows.ccorrica.pf.postprocessfilteropts", return_value=(args, prefilter)
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_io.readvectorsfromtextfile",
            return_value=(None, 0.0, None, tcdata, False, "text"),
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_resample.upsample",
            side_effect=lambda d, fi, fo, intfac=False: d,
        ),
        patch("rapidtide.workflows.ccorrica.tide_math.stdnormalize", side_effect=lambda x: x),
        patch(
            "rapidtide.workflows.ccorrica.tide_math.corrnormalize", side_effect=lambda x, **kw: x
        ),
        patch(
            "rapidtide.workflows.ccorrica.tide_corr.fastcorrelate", side_effect=mock_fastcorrelate
        ),
        patch("rapidtide.workflows.ccorrica.pearsonr", return_value=mock_pearsonr_result),
        patch(
            "rapidtide.workflows.ccorrica.tide_fit.findmaxlag_gauss",
            side_effect=mock_findmaxlag_gauss,
        ),
        patch("rapidtide.workflows.ccorrica.tide_stats.symmetrize", side_effect=mock_symmetrize),
        patch("rapidtide.workflows.ccorrica.nib.Nifti1Image", side_effect=mock_nifti1image),
        patch("rapidtide.workflows.ccorrica.tide_io.savetonifti", side_effect=stn_effect),
        patch("rapidtide.workflows.ccorrica.tide_io.writenpvecs", side_effect=wnp_effect),
    ):
        ccorrica(args)

    # Should have created headers for xcorr (4d) and pxcorr (4d)
    # plus one for 3d outputs (corrmax, corrlag, corrwidth, corrmask share same header)
    assert len(nifti_headers) >= 2
    expected_sampletime = 1.0 / 4.0
    for hdr in nifti_headers:
        assert abs(hdr["pixdim"][4] - expected_sampletime) < 1e-6

    if debug:
        print("test_ccorrica_nifti_pixdim_set passed")


# ---- integration test ----


def test_ccorrica_integration(debug=False):
    """Full integration test with multiple components and options."""
    args = _make_args(
        samplerate=2.0,
        oversampfactor=1,
        detrendorder=1,
        windowfunc="hamming",
        corrweighting="phat",
    )
    numcomponents = 4
    tclen = 100
    result = _run_ccorrica(args, numcomponents=numcomponents, tclen=tclen)

    # Verify all output files exist
    stn_files = [e[1] for e in result["savetonifti"]]
    wnp_files = [e[1] for e in result["writenpvecs"]]

    assert len(stn_files) == 6  # xcorr, pxcorr, corrmax, corrlag, corrwidth, corrmask
    assert len(wnp_files) == 6  # filtereddata, corrmax, corrlag, corrwidth, corrmask, reformdata

    # Verify shapes of matrix outputs
    for data, fname in result["writenpvecs"]:
        if "_corrmax.txt" in fname:
            assert data.shape == (numcomponents, numcomponents)
        if "_corrlag.txt" in fname:
            assert data.shape == (numcomponents, numcomponents)
        if "_corrwidth.txt" in fname:
            assert data.shape == (numcomponents, numcomponents)
        if "_corrmask.txt" in fname:
            assert data.shape == (numcomponents, numcomponents)
        if "_filtereddata.txt" in fname:
            assert data.shape == (numcomponents, tclen)
        if "_reformdata.txt" in fname:
            assert data.shape == (numcomponents, tclen)

    if debug:
        print("test_ccorrica_integration passed")


# ---- main test function ----


def test_ccorrica(debug=False):
    """Run all ccorrica tests."""
    # parser tests
    test_parser_required_args(debug=debug)
    test_parser_defaults(debug=debug)
    test_parser_samplerate(debug=debug)
    test_parser_sampletstep(debug=debug)
    test_parser_mutually_exclusive_rate(debug=debug)
    test_parser_corrweighting_choices(debug=debug)
    test_parser_detrendorder(debug=debug)
    test_parser_oversampfactor(debug=debug)
    test_parser_debug_flag(debug=debug)

    # ccorrica happy path tests
    test_ccorrica_basic_run(debug=debug)
    test_ccorrica_output_files(debug=debug)
    test_ccorrica_output_shapes(debug=debug)
    test_ccorrica_filtereddata_shape(debug=debug)
    test_ccorrica_reformdata_shape(debug=debug)
    test_ccorrica_single_component(debug=debug)
    test_ccorrica_two_components(debug=debug)
    test_ccorrica_xcorr_nifti_shape(debug=debug)

    # sample rate handling
    test_ccorrica_samplerate_from_file(debug=debug)
    test_ccorrica_samplerate_from_args(debug=debug)
    test_ccorrica_missing_samplerate_exits(debug=debug)

    # oversampling tests
    test_ccorrica_oversample(debug=debug)
    test_ccorrica_auto_oversampfactor(debug=debug)

    # option variation tests
    test_ccorrica_corrweighting_none(debug=debug)
    test_ccorrica_detrendorder_zero(debug=debug)
    test_ccorrica_debug_mode(debug=debug)
    test_ccorrica_window_func_hann(debug=debug)

    # symmetrization tests
    test_ccorrica_symmetrize_called(debug=debug)

    # correlation computation tests
    test_ccorrica_fastcorrelate_call_count(debug=debug)
    test_ccorrica_pearsonr_call_count(debug=debug)
    test_ccorrica_findmaxlag_gauss_called(debug=debug)

    # nifti header tests
    test_ccorrica_nifti_pixdim_set(debug=debug)

    # integration test
    test_ccorrica_integration(debug=debug)


if __name__ == "__main__":
    test_ccorrica(debug=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
from typing import Any

import numpy as np
from numpy.typing import NDArray

import rapidtide.io as tide_io


class fMRIData:
    header = None
    data = None
    data_byvoxel = None
    data_valid = None
    nativeshape = None
    mask = None
    validvoxels = None
    filename = None
    filetype = None

    def __init__(
        self,
        filename: str | None = None,
        header: Any | None = None,
        data: NDArray | None = None,
        mask: NDArray | None = None,
        validvoxels: NDArray | None = None,
        dims: tuple | None = None,
        sizes: tuple | None = None,
        description: str | None = None,
    ) -> None:
        """
        Initialize a new instance.

        Parameters
        ----------
        filename : str, optional
            Path to the data file. Default is None.
        header : Any, optional
            Header information associated with the data. Default is None.
        data : NDArray, optional
            Main data array. Default is None.
        mask : NDArray, optional
            Mask array for data filtering. Default is None.
        validvoxels : NDArray, optional
            Array indicating valid voxels. Default is None.
        dims : tuple, optional
            Dimensions of the data. Default is None.
        sizes : tuple, optional
            Size information for the data. Default is None.
        description : str, optional
            Description of the data. Default is None.

        Returns
        -------
        None
            This method initializes the instance attributes but does not return anything.

        Notes
        -----
        This constructor sets up the basic attributes for data handling. All parameters
        are optional and can be None, allowing for flexible initialization of instances.

        Examples
        --------
        >>> instance = MyClass()
        >>> instance = MyClass(filename="data.nii", data=np.array([1, 2, 3]))
        """
        self.filename = filename
        self.header = header
        self.data = data
        self.mask = mask
        self.validvoxels = validvoxels
        self.dims = dims
        self.sizes = sizes
        self.description = description

    def load(self, filename: str) -> None:
        """
        Load fMRI data from a file.

        Load fMRI data from a file, supporting text, CIFTI, and NIFTI formats.
        The function determines the file type and reads the data accordingly,
        setting appropriate attributes such as data shape, timepoints, and spatial dimensions.

        Parameters
        ----------
        filename : str
            Path to the input file. Supported formats are text, CIFTI, and NIFTI.

        Returns
        -------
        None
            This function does not return a value but updates the instance attributes
            with loaded data and metadata.

        Notes
        -----
        This function modifies the instance state by setting the following attributes:
        - `filename`: The path to the input file.
        - `filetype`: The type of the file ('text', 'cifti', or 'nifti').
        - `data`: The loaded data array.
        - `header`: The header information (if available).
        - `xsize`, `ysize`, `numslices`, `timepoints`: Spatial and temporal dimensions.
        - `thesizes`: A list containing the dimensions of the data.
        - `numspatiallocs`: Total number of spatial locations.
        - `nativespaceshape`: Shape of the native space.
        - `cifti_hdr`: CIFTI header (if applicable).

        Examples
        --------
        >>> loader = DataReader()
        >>> loader.load('fmri_data.nii.gz')
        >>> print(loader.timepoints)
        100
        """
        self.filename = filename
        ####################################################
        #  Read data
        ####################################################
        # open the fmri datafile
        if tide_io.checkiftext(self.filename):
            self.filetype = "text"
            self.data = tide_io.readvecs(self.filename)
            self.header = None
            theshape = np.shape(nim_data)
            self.xsize = theshape[0]
            self.ysize = 1
            self.numslices = 1
            self.timepoints = theshape[1]
            self.thesizes = [0, int(self.xsize), 1, 1, int(self.timepoints)]
            self.numspatiallocs = int(self.xsize)
            self.nativespaceshape = self.xsize
            self.cifti_hdr = None
        elif tide_io.checkifcifti(self.filename):
            self.filetype = "cifti"
            (
                cifti,
                cifti_hdr,
                self.data,
                self.header,
                thedims,
                thesizes,
                dummy,
            ) = tide_io.readfromcifti(self.filename)
            self.isgrayordinate = True
            self.timepoints = nim_data.shape[1]
            numspatiallocs = nim_data.shape[0]
            LGR.debug(f"cifti file has {timepoints} timepoints, {numspatiallocs} numspatiallocs")
            slicesize = numspatiallocs
            nativespaceshape = (1, 1, 1, 1, numspatiallocs)
        else:
            self.filetype = "nifti"
            LGR.debug("input file is NIFTI")
            nim, self.data, self.header, thedims, thesizes = tide_io.readfromnifti(fmrifilename)
            optiondict["isgrayordinate"] = False
            xsize, ysize, numslices, timepoints = tide_io.parseniftidims(thedims)
            numspatiallocs = int(xsize) * int(ysize) * int(numslices)
            cifti_hdr = None
            nativespaceshape = (xsize, ysize, numslices)
            xdim, ydim, slicethickness, tr = tide_io.parseniftisizes(thesizes)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
"""
Tests for rapidtide.simFuncClasses module.

This module tests the SimilarityFunctionator, MutualInformationator, Correlator,
SimilarityFunctionFitter, and FrequencyTracker classes.
"""

import numpy as np
import pytest

import rapidtide.filter as tide_filt
import rapidtide.simFuncClasses as tide_simfunc

# ============================================================================
# SimilarityFunctionator tests
# ============================================================================


class TestSimilarityFunctionator:
    """Tests for the SimilarityFunctionator base class."""

    @pytest.fixture
    def sample_filter(self):
        """Create a sample filter for testing."""
        return tide_filt.NoncausalFilter(filtertype="lfo")

    @pytest.fixture
    def sample_timecourse(self):
        """Create a sample timecourse for testing."""
        # Use low sampling frequency and long data for LFO filter compatibility
        t = np.linspace(0, 100, 200)
        return np.sin(2 * np.pi * 0.05 * t) + 0.5 * np.sin(2 * np.pi * 0.1 * t)

    def test_init_basic(self):
        """Test basic initialization of SimilarityFunctionator."""
        sim_func = tide_simfunc.SimilarityFunctionator(Fs=2.0)

        assert sim_func.Fs == 2.0
        assert sim_func.similarityfuncorigin == 0
        assert sim_func.lagmininpts == 0
        assert sim_func.lagmaxinpts == 0
        assert sim_func.detrendorder == 1
        assert sim_func.filterinputdata is True
        assert sim_func.debug is False

    def test_init_with_params(self, sample_filter):
        """Test initialization with custom parameters."""
        sim_func = tide_simfunc.SimilarityFunctionator(
            Fs=2.0,
            similarityfuncorigin=50,
            lagmininpts=10,
            lagmaxinpts=20,
            ncprefilter=sample_filter,
            negativegradient=True,
            detrendorder=2,
            filterinputdata=False,
            debug=True,
        )

        assert sim_func.Fs == 2.0
        assert sim_func.similarityfuncorigin == 50
        assert sim_func.lagmininpts == 10
        assert sim_func.lagmaxinpts == 20
        assert sim_func.negativegradient is True
        assert sim_func.detrendorder == 2
        assert sim_func.filterinputdata is False
        assert sim_func.debug is True

    def test_setFs(self):
        """Test setFs method."""
        sim_func = tide_simfunc.SimilarityFunctionator()

        sim_func.setFs(100.0)

        assert sim_func.Fs == 100.0

    def test_trim(self):
        """Test trim method."""
        sim_func = tide_simfunc.SimilarityFunctionator(
            similarityfuncorigin=50, lagmininpts=10, lagmaxinpts=20
        )

        vector = np.arange(100)
        trimmed = sim_func.trim(vector)

        # Should get elements from 50-10=40 to 50+20=70
        assert len(trimmed) == 30
        assert trimmed[0] == 40
        assert trimmed[-1] == 69

    def test_getfunction_no_data(self, capsys):
        """Test getfunction when no data is calculated."""
        sim_func = tide_simfunc.SimilarityFunctionator()

        result = sim_func.getfunction()

        assert result == (None, None, None)
        captured = capsys.readouterr()
        assert "must calculate similarity function" in captured.out


# ============================================================================
# MutualInformationator tests
# ============================================================================


class TestMutualInformationator:
    """Tests for the MutualInformationator class."""

    @pytest.fixture
    def sample_filter(self):
        """Create a sample filter for testing."""
        return tide_filt.NoncausalFilter(filtertype="lfo")

    @pytest.fixture
    def sample_timecourse(self):
        """Create a sample timecourse for testing."""
        t = np.linspace(0, 100, 200)
        return np.sin(2 * np.pi * 0.05 * t)

    def test_init_basic(self):
        """Test basic initialization of MutualInformationator."""
        mi = tide_simfunc.MutualInformationator(Fs=2.0)

        assert mi.Fs == 2.0
        assert mi.windowfunc == "hamming"
        assert mi.norm is True
        assert mi.madnorm is False
        assert mi.bins == 20
        assert mi.sigma == 0.25
        assert mi.smoothingtime == -1.0
        assert mi.mi_norm == 1.0

    def test_init_with_smoothing(self):
        """Test initialization with smoothing enabled."""
        mi = tide_simfunc.MutualInformationator(Fs=2.0, smoothingtime=2.0)

        assert mi.smoothingtime == 2.0
        assert mi.smoothingfilter is not None

    def test_init_with_custom_params(self, sample_filter):
        """Test initialization with custom parameters."""
        mi = tide_simfunc.MutualInformationator(
            Fs=2.0,
            windowfunc="hanning",
            norm=False,
            madnorm=True,
            bins=30,
            sigma=0.5,
            ncprefilter=sample_filter,
        )

        assert mi.windowfunc == "hanning"
        assert mi.norm is False
        assert mi.madnorm is True
        assert mi.bins == 30
        assert mi.sigma == 0.5

    def test_setbins(self):
        """Test setbins method."""
        mi = tide_simfunc.MutualInformationator(Fs=2.0)

        mi.setbins(50)

        assert mi.bins == 50

    def test_getnormfac(self):
        """Test getnormfac method."""
        mi = tide_simfunc.MutualInformationator(Fs=2.0)
        mi.mi_norm = 2.5

        assert mi.getnormfac() == 2.5

    def test_setreftc(self, sample_filter, sample_timecourse):
        """Test setreftc method."""
        mi = tide_simfunc.MutualInformationator(
            Fs=2.0,
            ncprefilter=sample_filter,
            lagmininpts=10,
            lagmaxinpts=10,
        )

        mi.setreftc(sample_timecourse, offset=0.0)

        assert mi.reftc is not None
        assert mi.prepreftc is not None
        assert mi.timeaxis is not None
        assert mi.timeaxisvalid is True
        assert mi.datavalid is False
        assert mi.similarityfunclen > 0

    def test_setlimits(self, sample_filter, sample_timecourse):
        """Test setlimits method."""
        mi = tide_simfunc.MutualInformationator(
            Fs=2.0,
            ncprefilter=sample_filter,
            lagmininpts=10,
            lagmaxinpts=10,
        )
        mi.setreftc(sample_timecourse)

        mi.setlimits(5, 15)

        assert mi.lagmininpts == 5
        assert mi.lagmaxinpts == 15

    def test_run(self, sample_filter, sample_timecourse):
        """Test run method."""
        mi = tide_simfunc.MutualInformationator(
            Fs=2.0,
            ncprefilter=sample_filter,
            lagmininpts=10,
            lagmaxinpts=10,
        )
        mi.setreftc(sample_timecourse)

        result = mi.run(sample_timecourse, trim=True)

        assert len(result) == 3
        simfunc, timeaxis, globalmax = result
        assert simfunc is not None
        assert timeaxis is not None
        assert globalmax >= 0
        assert mi.datavalid is True


# ============================================================================
# Correlator tests
# ============================================================================


class TestCorrelator:
    """Tests for the Correlator class."""

    @pytest.fixture
    def sample_filter(self):
        """Create a sample filter for testing."""
        return tide_filt.NoncausalFilter(filtertype="lfo")

    @pytest.fixture
    def sample_timecourse(self):
        """Create a sample timecourse for testing."""
        t = np.linspace(0, 100, 200)
        return np.sin(2 * np.pi * 0.05 * t)

    def test_init_basic(self):
        """Test basic initialization of Correlator."""
        corr = tide_simfunc.Correlator(Fs=2.0)

        assert corr.Fs == 2.0
        assert corr.windowfunc == "hamming"
        assert corr.corrweighting == "None"
        assert corr.corrpadding == 0
        assert corr.baselinefilter is None

    def test_init_with_custom_params(self, sample_filter):
        """Test initialization with custom parameters."""
        corr = tide_simfunc.Correlator(
            Fs=2.0,
            windowfunc="hanning",
            corrweighting="phat",
            corrpadding=100,
            ncprefilter=sample_filter,
        )

        assert corr.windowfunc == "hanning"
        assert corr.corrweighting == "phat"
        assert corr.corrpadding == 100

    def test_setlimits(self):
        """Test setlimits method."""
        corr = tide_simfunc.Correlator(Fs=2.0)

        corr.setlimits(5, 15)

        assert corr.lagmininpts == 5
        assert corr.lagmaxinpts == 15

    def test_setreftc(self, sample_filter, sample_timecourse):
        """Test setreftc method."""
        corr = tide_simfunc.Correlator(Fs=2.0, ncprefilter=sample_filter)

        corr.setreftc(sample_timecourse, offset=0.0)

        assert corr.reftc is not None
        assert corr.prepreftc is not None
        assert corr.timeaxis is not None
        assert corr.timeaxisvalid is True
        assert corr.datavalid is False
        assert corr.similarityfunclen == len(sample_timecourse) * 2 - 1

    def test_run(self, sample_filter, sample_timecourse):
        """Test run method."""
        corr = tide_simfunc.Correlator(
            Fs=2.0, ncprefilter=sample_filter, lagmininpts=10, lagmaxinpts=10
        )
        corr.setreftc(sample_timecourse)

        result = corr.run(sample_timecourse, trim=True)

        assert len(result) == 3
        simfunc, timeaxis, globalmax = result
        assert simfunc is not None
        assert timeaxis is not None
        assert globalmax >= 0
        assert corr.datavalid is True

    def test_run_no_trim(self, sample_filter, sample_timecourse):
        """Test run method without trimming."""
        corr = tide_simfunc.Correlator(
            Fs=2.0, ncprefilter=sample_filter, lagmininpts=10, lagmaxinpts=10
        )
        corr.setreftc(sample_timecourse)

        result = corr.run(sample_timecourse, trim=False)

        simfunc, timeaxis, globalmax = result
        assert len(simfunc) == corr.similarityfunclen

    def test_autocorrelation_peak(self, sample_filter, sample_timecourse):
        """Test that autocorrelation has peak at zero lag."""
        corr = tide_simfunc.Correlator(
            Fs=2.0, ncprefilter=sample_filter, lagmininpts=50, lagmaxinpts=50
        )
        corr.setreftc(sample_timecourse)

        simfunc, timeaxis, globalmax = corr.run(sample_timecourse, trim=False)

        # Autocorrelation should peak at zero lag (middle of the array)
        peak_idx = np.argmax(simfunc)
        peak_time = timeaxis[peak_idx]
        assert np.abs(peak_time) < 1.0  # Peak should be near zero


# ============================================================================
# SimilarityFunctionFitter tests
# ============================================================================


class TestSimilarityFunctionFitter:
    """Tests for the SimilarityFunctionFitter class."""

    @pytest.fixture
    def sample_corrtimeaxis(self):
        """Create a sample correlation time axis."""
        return np.linspace(-10, 10, 201)

    @pytest.fixture
    def sample_gaussian_corrfunc(self, sample_corrtimeaxis):
        """Create a sample Gaussian-shaped correlation function."""
        # Gaussian centered at lag=2 with amplitude 0.8 and sigma=1.5
        return 0.8 * np.exp(-((sample_corrtimeaxis - 2.0) ** 2) / (2 * 1.5**2))

    def test_init_basic(self):
        """Test basic initialization of SimilarityFunctionFitter."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        assert fitter.lagmin == -30.0
        assert fitter.lagmax == 30.0
        assert fitter.absmaxsigma == 1000.0
        assert fitter.absminsigma == 0.25
        assert fitter.hardlimit is True
        assert fitter.bipolar is False
        assert fitter.lthreshval == 0.0
        assert fitter.uthreshval == 1.0
        assert fitter.debug is False
        assert fitter.functype == "correlation"
        assert fitter.peakfittype == "gauss"

    def test_init_with_corrtimeaxis(self, sample_corrtimeaxis):
        """Test initialization with correlation time axis."""
        fitter = tide_simfunc.SimilarityFunctionFitter(corrtimeaxis=sample_corrtimeaxis)

        assert fitter.corrtimeaxis is not None
        assert len(fitter.corrtimeaxis) == len(sample_corrtimeaxis)

    def test_init_with_custom_params(self, sample_corrtimeaxis):
        """Test initialization with custom parameters."""
        fitter = tide_simfunc.SimilarityFunctionFitter(
            corrtimeaxis=sample_corrtimeaxis,
            lagmin=-5.0,
            lagmax=5.0,
            absmaxsigma=10.0,
            absminsigma=0.1,
            hardlimit=False,
            bipolar=True,
            lthreshval=0.1,
            uthreshval=0.9,
            functype="mutualinfo",
            peakfittype="quad",
        )

        assert fitter.lagmin == -5.0
        assert fitter.lagmax == 5.0
        assert fitter.absmaxsigma == 10.0
        assert fitter.absminsigma == 0.1
        assert fitter.hardlimit is False
        assert fitter.bipolar is True
        assert fitter.lthreshval == 0.1
        assert fitter.uthreshval == 0.9
        assert fitter.functype == "mutualinfo"
        assert fitter.peakfittype == "quad"

    def test_setfunctype(self):
        """Test setfunctype method."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        fitter.setfunctype("mutualinfo")

        assert fitter.functype == "mutualinfo"

    def test_setpeakfittype(self):
        """Test setpeakfittype method."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        fitter.setpeakfittype("quad")

        assert fitter.peakfittype == "quad"

    def test_setrange(self):
        """Test setrange method."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        fitter.setrange(-5.0, 5.0)

        assert fitter.lagmin == -5.0
        assert fitter.lagmax == 5.0

    def test_setcorrtimeaxis(self, sample_corrtimeaxis):
        """Test setcorrtimeaxis method."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        fitter.setcorrtimeaxis(sample_corrtimeaxis)

        assert fitter.corrtimeaxis is not None
        assert len(fitter.corrtimeaxis) == len(sample_corrtimeaxis)

    def test_setcorrtimeaxis_none(self):
        """Test setcorrtimeaxis with None."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        fitter.setcorrtimeaxis(None)

        assert fitter.corrtimeaxis is None

    def test_setguess(self):
        """Test setguess method."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        fitter.setguess(True, 5.0)

        assert fitter.useguess is True
        assert fitter.maxguess == 5.0

    def test_setlthresh(self):
        """Test setlthresh method."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        fitter.setlthresh(0.2)

        assert fitter.lthreshval == 0.2

    def test_setuthresh(self):
        """Test setuthresh method."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        fitter.setuthresh(0.8)

        assert fitter.uthreshval == 0.8

    def test_diagnosefail_no_error(self):
        """Test diagnosefail with no error."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        result = fitter.diagnosefail(np.uint32(0))

        assert result == "No error"

    def test_diagnosefail_single_error(self):
        """Test diagnosefail with single error flag."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        result = fitter.diagnosefail(fitter.FML_INITAMPLOW)

        assert "Initial amplitude too low" in result

    def test_diagnosefail_multiple_errors(self):
        """Test diagnosefail with multiple error flags."""
        fitter = tide_simfunc.SimilarityFunctionFitter()

        failreason = fitter.FML_INITAMPLOW | fitter.FML_FITLAGLOW
        result = fitter.diagnosefail(failreason)

        assert "Initial amplitude too low" in result
        assert "Fit Lag too low" in result

    def test_fit_gaussian(self, sample_corrtimeaxis, sample_gaussian_corrfunc):
        """Test fit method with Gaussian peak."""
        fitter = tide_simfunc.SimilarityFunctionFitter(
            corrtimeaxis=sample_corrtimeaxis,
            lagmin=-5.0,
            lagmax=5.0,
            peakfittype="gauss",
        )

        result = fitter.fit(sample_gaussian_corrfunc)

        maxindex, maxlag, maxval, maxsigma, maskval, failreason, peakstart, peakend = result

        # Peak should be found near lag=2.0 with amplitude ~0.8
        assert np.abs(maxlag - 2.0) < 0.5
        assert np.abs(maxval - 0.8) < 0.2
        assert maskval == 1  # Successful fit

    def test_fit_fastquad(self, sample_corrtimeaxis, sample_gaussian_corrfunc):
        """Test fit method with fastquad peak fitting."""
        fitter = tide_simfunc.SimilarityFunctionFitter(
            corrtimeaxis=sample_corrtimeaxis,
            lagmin=-5.0,
            lagmax=5.0,
            peakfittype="fastquad",
        )

        result = fitter.fit(sample_gaussian_corrfunc)

        maxindex, maxlag, maxval, maxsigma, maskval, failreason, peakstart, peakend = result

        assert maxlag is not None
        assert maxval is not None

    def test_fit_quad(self, sample_corrtimeaxis, sample_gaussian_corrfunc):
        """Test fit method with quadratic peak fitting."""
        fitter = tide_simfunc.SimilarityFunctionFitter(
            corrtimeaxis=sample_corrtimeaxis,
            lagmin=-5.0,
            lagmax=5.0,
            peakfittype="quad",
        )

        result = fitter.fit(sample_gaussian_corrfunc)

        maxindex, maxlag, maxval, maxsigma, maskval, failreason, peakstart, peakend = result

        assert maxlag is not None
        assert np.abs(maxlag - 2.0) < 0.5

    def test_fit_COM(self, sample_corrtimeaxis, sample_gaussian_corrfunc):
        """Test fit method with center-of-mass peak fitting."""
        fitter = tide_simfunc.SimilarityFunctionFitter(
            corrtimeaxis=sample_corrtimeaxis,
            lagmin=-5.0,
            lagmax=5.0,
            peakfittype="COM",
        )

        result = fitter.fit(sample_gaussian_corrfunc)

        maxindex, maxlag, maxval, maxsigma, maskval, failreason, peakstart, peakend = result

        assert maxlag is not None

    def test_fit_none_type(self, sample_corrtimeaxis, sample_gaussian_corrfunc):
        """Test fit method with None peak fitting (no refinement)."""
        fitter = tide_simfunc.SimilarityFunctionFitter(
            corrtimeaxis=sample_corrtimeaxis,
            lagmin=-5.0,
            lagmax=5.0,
            peakfittype="None",
        )

        result = fitter.fit(sample_gaussian_corrfunc)

        maxindex, maxlag, maxval, maxsigma, maskval, failreason, peakstart, peakend = result

        assert maxlag is not None

    def test_fit_bipolar(self, sample_corrtimeaxis):
        """Test fit method with bipolar option."""
        # Create negative peak
        negative_corrfunc = -0.8 * np.exp(-((sample_corrtimeaxis - 2.0) ** 2) / (2 * 1.5**2))

        fitter = tide_simfunc.SimilarityFunctionFitter(
            corrtimeaxis=sample_corrtimeaxis,
            lagmin=-5.0,
            lagmax=5.0,
            bipolar=True,
            peakfittype="gauss",
        )

        result = fitter.fit(negative_corrfunc)

        maxindex, maxlag, maxval, maxsigma, maskval, failreason, peakstart, peakend = result

        # Should find the negative peak
        assert maxval < 0

    def test_maxindex_noedge(self, sample_corrtimeaxis, sample_gaussian_corrfunc):
        """Test _maxindex_noedge method."""
        fitter = tide_simfunc.SimilarityFunctionFitter(
            corrtimeaxis=sample_corrtimeaxis,
            lagmin=-5.0,
            lagmax=5.0,
        )

        maxindex, flipfac = fitter._maxindex_noedge(sample_gaussian_corrfunc)

        # Peak should be found at correct index
        expected_index = np.argmax(sample_gaussian_corrfunc)
        assert maxindex == expected_index
        assert flipfac == 1.0


# ============================================================================
# FrequencyTracker tests
# ============================================================================


class TestFrequencyTracker:
    """Tests for the FrequencyTracker class."""

    def test_init_basic(self):
        """Test basic initialization of FrequencyTracker."""
        tracker = tide_simfunc.FrequencyTracker()

        assert tracker.lowerlim == 0.1
        assert tracker.upperlim == 0.6
        assert tracker.nperseg == 32
        assert tracker.Q == 10.0
        assert tracker.debug is False
        assert tracker.nfft == 32

    def test_init_with_custom_params(self):
        """Test initialization with custom parameters."""
        tracker = tide_simfunc.FrequencyTracker(
            lowerlim=0.2, upperlim=0.8, nperseg=64, Q=20.0, debug=True
        )

        assert tracker.lowerlim == 0.2
        assert tracker.upperlim == 0.8
        assert tracker.nperseg == 64
        assert tracker.Q == 20.0
        assert tracker.debug is True
        assert tracker.nfft == 64

    def test_track_sinusoid(self):
        """Test track method with a simple sinusoid."""
        tracker = tide_simfunc.FrequencyTracker(
            lowerlim=0.1, upperlim=0.5, nperseg=64, debug=False
        )

        # Create a 100-second signal at 10 Hz sampling with 0.2 Hz frequency
        fs = 10.0
        t = np.arange(0, 100, 1 / fs)
        signal_freq = 0.2
        x = np.sin(2 * np.pi * signal_freq * t)

        times, peakfreqs = tracker.track(x, fs)

        assert len(times) == len(peakfreqs)
        assert len(times) > 0

        # Most detected frequencies should be close to 0.2 Hz
        valid_freqs = peakfreqs[peakfreqs > 0]
        if len(valid_freqs) > 0:
            mean_freq = np.mean(valid_freqs)
            assert np.abs(mean_freq - signal_freq) < 0.1

    def test_track_returns_arrays(self):
        """Test that track returns proper numpy arrays."""
        tracker = tide_simfunc.FrequencyTracker(nperseg=32)

        fs = 10.0
        x = np.random.randn(500)

        times, peakfreqs = tracker.track(x, fs)

        assert isinstance(times, np.ndarray)
        assert isinstance(peakfreqs, np.ndarray)

    def test_clean(self):
        """Test clean method."""
        tracker = tide_simfunc.FrequencyTracker(
            lowerlim=0.1, upperlim=0.5, nperseg=64, debug=False
        )

        # Use higher sampling rate and longer signal to ensure enough data for filtering
        fs = 50.0
        t = np.arange(0, 100, 1 / fs)
        signal_freq = 0.2
        x = np.sin(2 * np.pi * signal_freq * t) + 0.5 * np.random.randn(len(t))

        times, peakfreqs = tracker.track(x, fs)

        # Clean the signal
        cleaned = tracker.clean(x, fs, times, peakfreqs, numharmonics=1)

        assert len(cleaned) == len(x)
        assert isinstance(cleaned, np.ndarray)


# ============================================================================
# Integration tests
# ============================================================================


class TestIntegration:
    """Integration tests combining multiple classes."""

    def test_correlator_with_fitter(self):
        """Test using Correlator output with SimilarityFunctionFitter."""
        # Create test signals - use simple synthetic data without filter
        t = np.linspace(-50, 50, 201)
        # Create a synthetic correlation function (Gaussian peak at lag=2)
        simfunc = 0.8 * np.exp(-((t - 2.0) ** 2) / (2 * 3.0**2))

        # Fit the result
        fitter = tide_simfunc.SimilarityFunctionFitter(
            corrtimeaxis=t,
            lagmin=-25.0,
            lagmax=25.0,
            peakfittype="gauss",
        )

        result = fitter.fit(simfunc)
        maxindex, maxlag, maxval, maxsigma, maskval, failreason, peakstart, peakend = result

        # Should find a peak near lag=2 with amplitude ~0.8
        assert maxval > 0.5
        assert np.abs(maxlag - 2.0) < 1.0

    def test_fitter_error_codes(self):
        """Test that fitter properly reports error conditions."""
        timeaxis = np.linspace(-10, 10, 201)
        # Very weak correlation
        weak_corr = 0.01 * np.exp(-((timeaxis - 2.0) ** 2) / (2 * 1.5**2))

        fitter = tide_simfunc.SimilarityFunctionFitter(
            corrtimeaxis=timeaxis,
            lagmin=-5.0,
            lagmax=5.0,
            lthreshval=0.1,  # Threshold higher than peak
            enforcethresh=True,
            peakfittype="gauss",
        )

        result = fitter.fit(weak_corr)
        maxindex, maxlag, maxval, maxsigma, maskval, failreason, peakstart, peakend = result

        # Should fail due to low amplitude
        diagnosis = fitter.diagnosefail(failreason)
        # Low amplitude fit or init should be detected
        assert failreason != fitter.FML_NOERROR or maskval == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import gc
import logging
from typing import Any, Callable

from numpy.typing import NDArray
from tqdm import tqdm

import rapidtide.multiproc as tide_multiproc


def run_multiproc(
    voxelfunc: Callable,
    packfunc: Callable,
    unpackfunc: Callable,
    voxelargs: list,
    voxelproducts: list,
    inputshape: tuple,
    voxelmask: NDArray,
    LGR: logging.Logger | None,
    nprocs: int,
    alwaysmultiproc: bool,
    showprogressbar: bool,
    chunksize: int,
    indexaxis: int = 0,
    procunit: str = "voxels",
    debug: bool = False,
    **kwargs: Any,
) -> int:
    """
    Execute voxel-wise processing with optional multiprocessing support.

    This function performs voxel-wise computations using the provided functions
    for processing, packing, and unpacking data. It supports both single-threaded
    and multi-process execution depending on the input parameters.

    Parameters
    ----------
    voxelfunc : callable
        Function to be applied to each voxel. It should accept a voxel index,
        packed arguments, and optional debug and keyword arguments.
    packfunc : callable
        Function to pack voxel arguments for use in `voxelfunc`. It takes a voxel
        index and `voxelargs` as input.
    unpackfunc : callable
        Function to unpack results from `voxelfunc` and store them in `voxelproducts`.
    voxelargs : list
        List of arguments to be passed to `packfunc`.
    voxelproducts : list
        List of arrays or data structures to be updated with results from `unpackfunc`.
    inputshape : tuple
        Shape of the input data, used to determine the number of voxels to process.
    voxelmask : ndarray
        Boolean or integer mask indicating which voxels to process.
    LGR : logging.Logger or None
        Logger instance for logging messages. If None, no logging is performed.
    nprocs : int
        Number of processes to use for multiprocessing. If 1, single-threaded execution is used.
    alwaysmultiproc : bool
        If True, forces multiprocessing even when `nprocs` is 1.
    showprogressbar : bool
        If True, displays a progress bar during processing.
    chunksize : int
        Size of chunks to process in each step for multiprocessing.
    indexaxis : int, optional
        Axis along which to iterate over voxels, default is 0.
    procunit : str, optional
        Unit of processing for progress bar, default is "voxels".
    debug : bool, optional
        If True, prints debug information, default is False.
    **kwargs : dict
        Additional keyword arguments passed to `voxelfunc`.

    Returns
    -------
    int
        Total number of voxels processed.

    Notes
    -----
    - If `nprocs` > 1 or `alwaysmultiproc` is True, multiprocessing is used.
    - Otherwise, a single-threaded loop is used with optional progress bar.
    - The function uses `tide_multiproc.run_multiproc` internally for multiprocessing.
    - Garbage collection is performed after processing.

    Examples
    --------
    >>> run_multiproc(
    ...     voxelfunc=my_voxel_func,
    ...     packfunc=pack_voxel_args,
    ...     unpackfunc=unpack_results,
    ...     voxelargs=[arg1, arg2],
    ...     voxelproducts=[output_array],
    ...     inputshape=(100, 100, 100),
    ...     voxelmask=mask_array,
    ...     LGR=None,
    ...     nprocs=4,
    ...     alwaysmultiproc=False,
    ...     showprogressbar=True,
    ...     chunksize=10,
    ... )
    1000
    """
    if debug:
        print(f"{len(voxelproducts)=}, {voxelproducts[0].shape}")
    if nprocs > 1 or alwaysmultiproc:
        # define the consumer function here so it inherits most of the arguments
        def theconsumerfunc(inQ, outQ):
            while True:
                try:
                    # get a new message
                    val = inQ.get()

                    # this is the 'TERM' signal
                    if val is None:
                        break

                    # process and send the data
                    outQ.put(
                        voxelfunc(
                            val,
                            packfunc(val, voxelargs),
                            debug=debug,
                            **kwargs,
                        )
                    )
                except Exception as e:
                    print("error!", e)
                    break

        data_out = tide_multiproc.run_multiproc(
            theconsumerfunc,
            inputshape,
            voxelmask,
            indexaxis=indexaxis,
            procunit=procunit,
            verbose=(LGR is not None),
            nprocs=nprocs,
            showprogressbar=showprogressbar,
            chunksize=chunksize,
        )

        # unpack the data
        volumetotal = 0
        for returnvals in data_out:
            volumetotal += 1
            unpackfunc(returnvals, voxelproducts)
        del data_out
    else:
        volumetotal = 0
        for vox in tqdm(
            range(0, inputshape[indexaxis]),
            desc="Voxel",
            unit=procunit,
            disable=(not showprogressbar),
        ):
            if voxelmask[vox] > 0:
                dothisone = True
            else:
                dothisone = False
            if dothisone:
                returnvals = voxelfunc(
                    vox,
                    packfunc(vox, voxelargs),
                    debug=debug,
                    **kwargs,
                )
                unpackfunc(returnvals, voxelproducts)
                volumetotal += 1

    # garbage collect
    uncollected = gc.collect()
    if uncollected != 0:
        if LGR is not None:
            LGR.info(f"garbage collected - unable to collect {uncollected} objects")
    else:
        if LGR is not None:
            LGR.info("garbage collected")

    return volumetotal



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2024-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
from argparse import Namespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd

import rapidtide.io as tide_io
import rapidtide.stats as tide_stats


def _get_parser() -> Any:
    """
    Argument parser for mergequality.

    Creates and configures an argument parser for the mergequality tool that merges
    rapidtide quality check data from multiple runs.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser object with all required and optional arguments
        for the mergequality tool.

    Notes
    -----
    The returned parser is configured with:
    - Required arguments: --input and --outputroot
    - Optional arguments: --keyfile, --showhists, --addgraymetrics, --addwhitemetrics, --debug

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args(['--input', 'run1.csv', 'run2.csv', '--outputroot', 'merged'])
    """
    parser = argparse.ArgumentParser(
        prog="mergequality",
        description=("Merge rapidtide quality check data from several runs."),
        allow_abbrev=False,
    )

    # Required arguments
    parser.add_argument("--input", action="store", type=str, nargs="*", required=True)
    parser.add_argument("--outputroot", action="store", type=str, required=True)

    # add optional
    parser.add_argument("--keyfile", action="store", type=str, default=None)
    parser.add_argument(
        "--showhists",
        action="store_true",
        help=("Display the histograms of the tracked quantities."),
        default=False,
    )
    parser.add_argument(
        "--addgraymetrics",
        action="store_true",
        help=("Include gray matter only metrics."),
        default=False,
    )
    parser.add_argument(
        "--addwhitemetrics",
        action="store_true",
        help=("Include white matter only metrics."),
        default=False,
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help=("Output additional debugging information."),
        default=False,
    )

    return parser


def mergequality(args: Any) -> None:
    """
    Merge quality metrics from multiple input JSON files into a single CSV file.

    This function reads quality metrics from input JSON files and combines them
    into a structured DataFrame. It supports optional inclusion of gray and white
    matter specific metrics based on command-line arguments. Histograms for each
    metric are also generated and saved.

    Parameters
    ----------
    args : Any
        An object containing the following attributes:
        - input : list of str
            List of input JSON file paths to process.
        - outputroot : str
            Base name for output CSV and histogram files.
        - keyfile : str, optional
            Path to a JSON file containing key metrics definitions. If None,
            default metrics are used.
        - addgraymetrics : bool
            Whether to include gray matter specific metrics.
        - addwhitemetrics : bool
            Whether to include white matter specific metrics.
        - showhists : bool
            Whether to display histograms.
        - debug : bool
            Whether to print debug information.

    Returns
    -------
    None
        This function does not return a value but saves a CSV file and histogram
        plots to disk.

    Notes
    -----
    The function uses `tide_io.readdictfromjson` to read input JSON files and
    `tide_stats.makeandsavehistogram` to generate histograms. Default metrics
    are included for mask, regressor, lag, laggrad, strength, and MTT.
    Gray and white matter metrics are conditionally added based on `args`.

    Examples
    --------
    >>> args = type('Args', (), {
    ...     'input': ['file1.json', 'file2.json'],
    ...     'outputroot': 'output',
    ...     'keyfile': None,
    ...     'addgraymetrics': True,
    ...     'addwhitemetrics': False,
    ...     'showhists': True,
    ...     'debug': False
    ... })()
    >>> mergequality(args)
    """
    if args.debug:
        print(f"{args.input=}")
        print(f"{args.outputroot=}")
        print(f"{args.keyfile=}")

    if args.keyfile is not None:
        thekeydict = tide_io.readdictfromjson(args.keyfile)
    else:
        thekeydict = {
            "mask": {
                "lagmaskvoxels": 215891,
                "meanrelsize": 4.180947793099295,
                "p_lt_0p001relsize": 0.6806073435205728,
                "p_lt_0p005relsize": 0.745116748729683,
                "p_lt_0p010relsize": 0.7726074732156505,
                "p_lt_0p050relsize": 0.838487940673766,
                "refinerelsize": 0.5496616348064532,
            },
            "regressor": {
                "first_kurtosis": 2.911833576023774,
                "first_kurtosis_p": 1.143835967291008e-38,
                "first_kurtosis_z": 13.00514376674551,
                "first_skewness": -0.3219158734451699,
                "first_skewness_p": 2.9429479496726994e-10,
                "first_skewness_z": -6.301754636834285,
                "first_spectralflatness": 0.9306835675389035,
                "last_kurtosis": 2.6687828914360496,
                "last_kurtosis_p": 1.2610242749531053e-35,
                "last_kurtosis_z": 12.458253243613367,
                "last_skewness": 0.06902106526612979,
                "last_skewness_p": 0.1667298705677166,
                "last_skewness_z": 1.3827880292121477,
                "last_spectralflatness": 0.9214471595162257,
            },
            "lag": {
                "centerofmass": 0.18365522576293233,
                "kurtosis": 2.6365118239988172,
                "kurtosis_p": 0.0,
                "kurtosis_z": 115.68766626074614,
                "pct02": -3.5552243631332305,
                "pct25": -1.114490891184175,
                "pct50": -0.1257154777049373,
                "pct75": 1.0872777685344683,
                "pct98": 5.428974111640412,
                "q1width": -3.5552243631332305,
                "q2width": -1.114490891184175,
                "q3width": -0.1257154777049373,
                "q4width": 1.0872777685344683,
                "mid50width": 5.428974111640412,
                "peakheight": 0.268346902434662,
                "peakloc": -0.32178217821782207,
                "peakpercentile": 44.776762347666185,
                "peakwidth": 3.2673267326732676,
                "skewness": 0.8500928388366628,
                "skewness_p": 0.0,
                "skewness_z": 140.97401621947097,
                "voxelsincluded": 215891,
            },
            "laggrad": {
                "centerofmass": 0.571792758578445,
                "kurtosis": 9.628956515105777,
                "kurtosis_p": 0.0,
                "kurtosis_z": 178.87307892867446,
                "pct02": 0.08644029467314696,
                "pct25": 0.2526212777361018,
                "pct50": 0.41670054269474405,
                "pct75": 0.7196355014183266,
                "pct98": 2.445977359086096,
                "q1width": -3.5552243631332305,
                "q2width": -1.114490891184175,
                "q3width": -0.1257154777049373,
                "q4width": 1.0872777685344683,
                "mid50width": 5.428974111640412,
                "peakheight": 1.7495682765313678,
                "peakloc": 0.2524752475247525,
                "peakpercentile": 24.97466346179356,
                "peakwidth": 0.6534653465346536,
                "skewness": 2.6454479455818714,
                "skewness_p": 0.0,
                "skewness_z": 273.76650916159275,
                "voxelsincluded": 188463,
            },
            "strength": {
                "centerofmass": 0.3718351617952105,
                "kurtosis": -0.7116698199891451,
                "kurtosis_p": 0.0,
                "kurtosis_z": -109.8631499714451,
                "pct02": 0.0463983482697488,
                "pct25": 0.2195563043221062,
                "pct50": 0.35605876676999876,
                "pct75": 0.5108963405705843,
                "pct98": 0.7675960511390326,
                "q1width": -3.5552243631332305,
                "q2width": -1.114490891184175,
                "q3width": -0.1257154777049373,
                "q4width": 1.0872777685344683,
                "mid50width": 5.428974111640412,
                "peakheight": 1.9180975584901625,
                "peakloc": 0.2722772277227723,
                "peakpercentile": 34.364100402517934,
                "peakwidth": 0.7326732673267327,
                "skewness": 0.2767447291065046,
                "skewness_p": 0.0,
                "skewness_z": 51.540795677466,
                "voxelsincluded": 215891,
            },
            "MTT": {
                "centerofmass": 1.1854362572800032,
                "kurtosis": 6.48403822013551,
                "kurtosis_p": 0.0,
                "kurtosis_z": 169.01367062971818,
                "pct02": 0.0,
                "pct25": 0.0,
                "pct50": 0.0,
                "pct75": 0.9117823500153743,
                "pct98": 2.835547357601372,
                "q1width": -3.5552243631332305,
                "q2width": -1.114490891184175,
                "q3width": -0.1257154777049373,
                "q4width": 1.0872777685344683,
                "mid50width": 5.428974111640412,
                "peakheight": 0.33323899429333736,
                "peakloc": 0.8415841584158417,
                "peakpercentile": 39.06788519856343,
                "peakwidth": 1.386138613861386,
                "skewness": 2.0750539699859707,
                "skewness_p": 0.0,
                "skewness_z": 257.5035937509046,
                "voxelsincluded": 215891,
            },
        }

        thewhitematterdict = {
            "whiteonly-lag": {
                "centerofmass": 0.12863780691943305,
                "kurtosis": 6.051218980924041,
                "kurtosis_p": 0.0,
                "kurtosis_z": 97.74313967555645,
                "pct02": -3.447084496826285,
                "pct25": -1.153943456786113,
                "pct50": 0.40468708058815667,
                "pct75": 2.5020675667656325,
                "pct98": 35.65786745845405,
                "q1width": -3.5552243631332305,
                "q2width": -1.114490891184175,
                "q3width": -0.1257154777049373,
                "q4width": 1.0872777685344683,
                "mid50width": 5.428974111640412,
                "peakheight": 0.20154056103108517,
                "peakloc": -0.7673267326732676,
                "peakpercentile": 31.311237290373697,
                "peakwidth": 7.7227722772277225,
                "skewness": 2.583265856366246,
                "skewness_p": 0.0,
                "skewness_z": 171.4754157543282,
                "voxelsincluded": 75730,
            },
            "whiteonly-laggrad": {
                "centerofmass": 0.4400415073580781,
                "kurtosis": 4.8057153283315746,
                "kurtosis_p": 0.0,
                "kurtosis_z": 88.82511535725718,
                "pct02": 0.07577299562385723,
                "pct25": 0.2357562409434768,
                "pct50": 0.4080402253691755,
                "pct75": 0.9014255193901806,
                "pct98": 12.105429775628355,
                "q1width": -3.5552243631332305,
                "q2width": -1.114490891184175,
                "q3width": -0.1257154777049373,
                "q4width": 1.0872777685344683,
                "mid50width": 5.428974111640412,
                "peakheight": 2.137460662083183,
                "peakloc": 0.22277227722772275,
                "peakpercentile": 22.70194611181395,
                "peakwidth": 0.5346534653465347,
                "skewness": 2.3423104651022015,
                "skewness_p": 0.0,
                "skewness_z": 161.40631688712926,
                "voxelsincluded": 74302,
            },
            "whiteonly-strength": {
                "centerofmass": 0.4427689302668548,
                "kurtosis": -0.3709348443577771,
                "kurtosis_p": 1.4319828709989065e-146,
                "kurtosis_z": -25.781443647052853,
                "pct02": 0.062435511959463494,
                "pct25": 0.3373922124837407,
                "pct50": 0.4652726794639804,
                "pct75": 0.5657461961283083,
                "pct98": 0.7482491589950304,
                "q1width": -3.5552243631332305,
                "q2width": -1.114490891184175,
                "q3width": -0.1257154777049373,
                "q4width": 1.0872777685344683,
                "mid50width": 5.428974111640412,
                "peakheight": 2.6820414630925633,
                "peakloc": 0.4900990099009901,
                "peakpercentile": 56.33302522118051,
                "peakwidth": 0.3168316831683169,
                "skewness": -0.4091197098383096,
                "skewness_p": 0.0,
                "skewness_z": -44.23408729903814,
                "voxelsincluded": 75730,
            },
        }

        thegraymatterdict = {
            "grayonly-lag": {
                "centerofmass": 0.12863780691943305,
                "kurtosis": 6.051218980924041,
                "kurtosis_p": 0.0,
                "kurtosis_z": 97.74313967555645,
                "pct02": -3.447084496826285,
                "pct25": -1.153943456786113,
                "pct50": 0.40468708058815667,
                "pct75": 2.5020675667656325,
                "pct98": 35.65786745845405,
                "q1width": -3.5552243631332305,
                "q2width": -1.114490891184175,
                "q3width": -0.1257154777049373,
                "q4width": 1.0872777685344683,
                "mid50width": 5.428974111640412,
                "peakheight": 0.20154056103108517,
                "peakloc": -0.7673267326732676,
                "peakpercentile": 31.311237290373697,
                "peakwidth": 7.7227722772277225,
                "skewness": 2.583265856366246,
                "skewness_p": 0.0,
                "skewness_z": 171.4754157543282,
                "voxelsincluded": 75730,
            },
            "grayonly-laggrad": {
                "centerofmass": 0.4400415073580781,
                "kurtosis": 4.8057153283315746,
                "kurtosis_p": 0.0,
                "kurtosis_z": 88.82511535725718,
                "pct02": 0.07577299562385723,
                "pct25": 0.2357562409434768,
                "pct50": 0.4080402253691755,
                "pct75": 0.9014255193901806,
                "pct98": 12.105429775628355,
                "q1width": -3.5552243631332305,
                "q2width": -1.114490891184175,
                "q3width": -0.1257154777049373,
                "q4width": 1.0872777685344683,
                "mid50width": 5.428974111640412,
                "peakheight": 2.137460662083183,
                "peakloc": 0.22277227722772275,
                "peakpercentile": 22.70194611181395,
                "peakwidth": 0.5346534653465347,
                "skewness": 2.3423104651022015,
                "skewness_p": 0.0,
                "skewness_z": 161.40631688712926,
                "voxelsincluded": 74302,
            },
            "grayonly-strength": {
                "centerofmass": 0.4427689302668548,
                "kurtosis": -0.3709348443577771,
                "kurtosis_p": 1.4319828709989065e-146,
                "kurtosis_z": -25.781443647052853,
                "pct02": 0.062435511959463494,
                "pct25": 0.3373922124837407,
                "pct50": 0.4652726794639804,
                "pct75": 0.5657461961283083,
                "pct98": 0.7482491589950304,
                "q1width": -3.5552243631332305,
                "q2width": -1.114490891184175,
                "q3width": -0.1257154777049373,
                "q4width": 1.0872777685344683,
                "mid50width": 5.428974111640412,
                "peakheight": 2.6820414630925633,
                "peakloc": 0.4900990099009901,
                "peakpercentile": 56.33302522118051,
                "peakwidth": 0.3168316831683169,
                "skewness": -0.4091197098383096,
                "skewness_p": 0.0,
                "skewness_z": -44.23408729903814,
                "voxelsincluded": 75730,
            },
        }
        if args.addgraymetrics:
            thekeydict.update(thegraymatterdict)
        if args.addwhitemetrics:
            thekeydict.update(thewhitematterdict)

    thecolumns = ["datasource"]
    thedatadict = {"datasource": []}
    for key in thekeydict.keys():
        for subkey in thekeydict[key]:
            thecolumns.append(key + "_" + str(subkey))
            thedatadict[thecolumns[-1]] = []

    if args.debug:
        print(thecolumns)

    for theinput in args.input:
        inputdict = tide_io.readdictfromjson(theinput)
        thedatadict["datasource"].append(theinput)
        for column in thecolumns[1:]:
            keyparts = column.split("_")
            try:
                thedataitem = inputdict[keyparts[0]]["_".join(keyparts[1:])]
            except KeyError:
                thedataitem = None
            thedatadict[column].append(thedataitem)
    df = pd.DataFrame(thedatadict, columns=thecolumns)
    df.to_csv(args.outputroot + ".csv", index=False)

    for column in thecolumns[1:]:
        tide_stats.makeandsavehistogram(
            df[column].to_numpy(),
            51,
            0,
            args.outputroot + "_" + column,
            displaytitle=column,
            displayplots=args.showhists,
            normalize=True,
            append=False,
            debug=False,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import numpy as np

from rapidtide.workflows.adjustoffset import _get_parser as adjustoffset_getparser
from rapidtide.workflows.aligntcs import _get_parser as aligntcs_getparser
from rapidtide.workflows.atlasaverage import _get_parser as atlasaverage_getparser
from rapidtide.workflows.atlastool import _get_parser as atlastool_getparser
from rapidtide.workflows.calctexticc import _get_parser as calctexticc_getparser
from rapidtide.workflows.ccorrica import _get_parser as ccorrica_getparser
from rapidtide.workflows.delayvar import _get_parser as delayvar_getparser
from rapidtide.workflows.diffrois import _get_parser as diffrois_getparser
from rapidtide.workflows.endtidalproc import _get_parser as endtidalproc_getparser
from rapidtide.workflows.fdica import _get_parser as fdica_getparser
from rapidtide.workflows.filtnifti import _get_parser as filtnifti_getparser
from rapidtide.workflows.filttc import _get_parser as filttc_getparser
from rapidtide.workflows.fixtr import _get_parser as fixtr_getparser
from rapidtide.workflows.gmscalc import _get_parser as gmscalc_getparser
from rapidtide.workflows.happy2std import _get_parser as happy2std_getparser
from rapidtide.workflows.happy_parser import _get_parser as happy_parser_getparser
from rapidtide.workflows.histnifti import _get_parser as histnifti_getparser
from rapidtide.workflows.histtc import _get_parser as histtc_getparser
from rapidtide.workflows.linfitfilt import _get_parser as linfitfilt_getparser
from rapidtide.workflows.localflow import _get_parser as localflow_getparser
from rapidtide.workflows.mergequality import _get_parser as mergequality_getparser
from rapidtide.workflows.niftidecomp import (
    _get_parser_spatial as niftidecomp_getparser_spatial,
)
from rapidtide.workflows.niftidecomp import (
    _get_parser_temporal as niftidecomp_getparser_temporal,
)
from rapidtide.workflows.niftistats import _get_parser as niftistats_getparser
from rapidtide.workflows.pairproc import _get_parser as pairproc_getparser
from rapidtide.workflows.pairwisemergenifti import (
    _get_parser as pairwisemergenifti_getparser,
)
from rapidtide.workflows.physiofreq import _get_parser as physiofreq_getparser
from rapidtide.workflows.pixelcomp import _get_parser as pixelcomp_getparser
from rapidtide.workflows.plethquality import _get_parser as plethquality_getparser
from rapidtide.workflows.polyfitim import _get_parser as polyfitim_getparser
from rapidtide.workflows.proj2flow import _get_parser as proj2flow_getparser
from rapidtide.workflows.rankimage import _get_parser as rankimage_getparser
from rapidtide.workflows.rapidtide2std import _get_parser as rapidtide2std_getparser
from rapidtide.workflows.resamplenifti import _get_parser as resamplenifti_getparser
from rapidtide.workflows.resampletc import _get_parser as resampletc_getparser
from rapidtide.workflows.retrolagtcs import _get_parser as retrolagtcs_getparser
from rapidtide.workflows.retroregress import _get_parser as retroregress_getparser
from rapidtide.workflows.roisummarize import _get_parser as roisummarize_getparser
from rapidtide.workflows.runqualitycheck import _get_parser as runqualitycheck_getparser
from rapidtide.workflows.showarbcorr import _get_parser as showarbcorr_getparser
from rapidtide.workflows.showhist import _get_parser as showhist_getparser
from rapidtide.workflows.showstxcorr import _get_parser as showstxcorr_getparser
from rapidtide.workflows.showtc import _get_parser as showtc_getparser
from rapidtide.workflows.showxcorrx import _get_parser as showxcorrx_getparser
from rapidtide.workflows.showxy import _get_parser as showxy_getparser
from rapidtide.workflows.simdata import _get_parser as simdata_getparser
from rapidtide.workflows.spatialfit import _get_parser as spatialfit_getparser
from rapidtide.workflows.spatialmi import _get_parser as spatialmi_getparser
from rapidtide.workflows.spectrogram import _get_parser as spectrogram_getparser
from rapidtide.workflows.synthASL import _get_parser as synthASL_getparser
from rapidtide.workflows.tcfrom2col import _get_parser as tcfrom2col_getparser
from rapidtide.workflows.tcfrom3col import _get_parser as tcfrom3col_getparser
from rapidtide.workflows.variabilityizer import _get_parser as variabilityizer_getparser


def test_parsers(debug=False):
    parserlist = [
        adjustoffset_getparser,
        aligntcs_getparser,
        atlasaverage_getparser,
        atlastool_getparser,
        calctexticc_getparser,
        ccorrica_getparser,
        delayvar_getparser,
        diffrois_getparser,
        endtidalproc_getparser,
        fdica_getparser,
        filtnifti_getparser,
        filttc_getparser,
        fixtr_getparser,
        gmscalc_getparser,
        happy_parser_getparser,
        happy2std_getparser,
        histnifti_getparser,
        histtc_getparser,
        linfitfilt_getparser,
        localflow_getparser,
        mergequality_getparser,
        niftidecomp_getparser_temporal,
        niftidecomp_getparser_spatial,
        niftistats_getparser,
        pairproc_getparser,
        pairwisemergenifti_getparser,
        physiofreq_getparser,
        pixelcomp_getparser,
        plethquality_getparser,
        polyfitim_getparser,
        proj2flow_getparser,
        rankimage_getparser,
        rapidtide2std_getparser,
        resamplenifti_getparser,
        resampletc_getparser,
        retrolagtcs_getparser,
        retroregress_getparser,
        roisummarize_getparser,
        runqualitycheck_getparser,
        showarbcorr_getparser,
        showhist_getparser,
        showstxcorr_getparser,
        showtc_getparser,
        showxcorrx_getparser,
        showxy_getparser,
        simdata_getparser,
        spatialfit_getparser,
        spatialmi_getparser,
        spectrogram_getparser,
        synthASL_getparser,
        tcfrom2col_getparser,
        tcfrom3col_getparser,
        variabilityizer_getparser,
    ]
    try:
        import tensorflow as tf

        dlfilterloads = True
    except ImportError:
        dlfilterloads = False
    if dlfilterloads:
        from rapidtide.workflows.applydlfilter import (
            _get_parser as applydlfilter_getparser,
        )

        parserlist.append(applydlfilter_getparser)

    for thegetparser in parserlist:
        theusage = thegetparser().format_help()
        if debug:
            print(theusage)


if __name__ == "__main__":
    test_parsers(debug=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
from argparse import Namespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray

import rapidtide.filter as tide_filt
import rapidtide.io as tide_io


def _get_parser() -> Any:
    """
    Argument parser for filtnifti.

    Creates and configures an argument parser for the filtnifti command-line tool
    that temporally filters NIFTI files.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser object with required command-line arguments

    Notes
    -----
    The parser expects four command-line arguments in the following order:
    1. inputfilename - path to input NIFTI file
    2. outputfilename - path to output NIFTI file
    3. lowestfreq - low passband frequency limit in Hz (negative values disable HPF)
    4. highestfreq - high passband frequency limit in Hz (negative values disable LPF)

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args(['input.nii', 'output.nii', '0.01', '0.1'])
    """
    # get the command line parameters
    parser = argparse.ArgumentParser(
        prog="filtnifti",
        description="Temporally filters a NIFTI file.",
        allow_abbrev=False,
    )
    parser.add_argument("inputfilename", type=str, help="The name of the input nifti file.")
    parser.add_argument("outputfilename", type=str, help="The name of the output nifti file.")
    parser.add_argument(
        "lowestfreq",
        type=float,
        help="The low passband frequency limit in Hz (set less than zero to disable HPF).",
    )
    parser.add_argument(
        "highestfreq",
        type=float,
        help="The high passband frequency limit in Hz (set less than zero to disable LPF)",
    )
    return parser


def filtnifti(args: Any) -> None:
    """
    Apply a frequency-domain filter to a 4D NIfTI file containing fMRI data.

    This function reads an input NIfTI file, applies a bandpass or arbitrary frequency filter
    to each voxel's time series, and saves the filtered data to a new NIfTI file. The filter
    parameters are specified via the `args` object, which should contain attributes such as
    `inputfilename`, `outputfilename`, `lowestfreq`, and `highestfreq`.

    Parameters
    ----------
    args : Any
        An object containing the following attributes:
        - inputfilename : str
            Path to the input NIfTI file.
        - outputfilename : str
            Path to the output NIfTI file.
        - lowestfreq : float
            Lowest frequency for the filter. If negative, highpass filtering is disabled.
        - highestfreq : float
            Highest frequency for the filter. If negative, lowpass filtering is disabled.

    Returns
    -------
    None
        This function does not return a value but saves the filtered data to a NIfTI file.

    Notes
    -----
    - The function uses a non-causal filter (`tide_filt.NoncausalFilter`) for filtering.
    - The TR (repetition time) is read from the NIfTI header, and the sampling frequency
      is computed as `Fs = 1 / TR`.
    - Filtering is applied slice-by-slice to each voxel's time series.
    - If `lowestfreq` or `highestfreq` is set to a negative value, the corresponding filter
      is disabled.

    Examples
    --------
    Assuming `args` is an object with the required attributes:

    >>> filtnifti(args)

    This will read the input NIfTI file, apply the specified filter, and save the result
    to the output file.
    """
    # get the input TR
    inputtr_fromfile, numinputtrs = tide_io.fmritimeinfo(args.inputfilename)
    print("input data: ", numinputtrs, " timepoints, tr = ", inputtr_fromfile)

    # sanity check the filter frequencies
    ftype = "arb"
    if args.lowestfreq < 0.0:
        print("disabling highpass filter")
        args.lowestfreq = 0.0

    if args.highestfreq < 0.0:
        print("disabling lowpass filter")

    if ftype == "bandpass":
        print("passing frequencies between ", args.lowestfreq, " and ", args.highestfreq)

    input_img, input_data, input_hdr, thedims, thesizes = tide_io.readfromnifti(args.inputfilename)
    if input_hdr.get_xyzt_units()[1] == "msec":
        tr = thesizes[4] / 1000.0
    else:
        tr = thesizes[4]
    Fs = 1.0 / tr
    print("tr from header =", tr, ", sample frequency is ", Fs)

    xsize, ysize, numslices, timepoints = tide_io.parseniftidims(thedims)
    filteredtcs = np.zeros((xsize, ysize, numslices, timepoints), dtype="float")

    # cycle over all voxels
    print("now cycling over all voxels")
    theprefilter = tide_filt.NoncausalFilter()
    theprefilter.settype("arb")
    theprefilter.setfreqs(args.lowestfreq, args.lowestfreq, args.highestfreq, args.highestfreq)

    for zloc in range(numslices):
        print("processing slice ", zloc)
        for yloc in range(ysize):
            for xloc in range(xsize):
                filteredtcs[xloc, yloc, zloc, :] = theprefilter.apply(
                    Fs, input_data[xloc, yloc, zloc, :]
                )

    # now do the ones with other numbers of time points
    tide_io.savetonifti(filteredtcs, input_hdr, args.outputfilename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import os

import matplotlib as mpl
import numpy as np

import rapidtide.filter as tide_filt
import rapidtide.miscmath as tide_math
import rapidtide.resample as tide_resample
import rapidtide.simFuncClasses as tide_simFuncClasses
import rapidtide.workflows.cleanregressor as tide_cleanregressor
from rapidtide.tests.utils import get_examples_path, get_test_temp_path, mse


def test_cleanregressor(debug=False, local=False, displayplots=False):
    # set input and output directories
    exampleroot = get_examples_path(local)
    testtemproot = get_test_temp_path(local)

    outputname = os.path.join(testtemproot, "cleanregressortest")
    thepass = 1
    padtrs = 30
    fmrifreq = 1.0
    oversampfac = 2
    oversampfreq = oversampfac * fmrifreq
    theprefilter = tide_filt.NoncausalFilter("lfo")
    lagmin = -30
    lagmax = 30
    lagmininpts = int((lagmin * oversampfreq) - 0.5)
    lagmaxinpts = int((lagmax * oversampfreq) + 0.5)
    lagmod = 1000.0
    noiseamp = 0.25
    detrendorder = 3
    windowfunc = "hamming"

    tclen = 500
    osvalidsimcalcstart = 0
    osvalidsimcalcend = tclen * oversampfac

    theCorrelator = tide_simFuncClasses.Correlator(
        Fs=oversampfreq,
        ncprefilter=theprefilter,
        detrendorder=1,
        windowfunc="hamming",
        corrweighting="phat",
    )
    theFitter = tide_simFuncClasses.SimilarityFunctionFitter(
        lagmod=lagmod,
        lagmin=lagmin,
        lagmax=lagmax,
        debug=debug,
        enforcethresh=False,
        zerooutbadfit=False,
    )

    # make a reference timecourse
    rng = np.random.default_rng(seed=1234)
    basewave = theprefilter.apply(fmrifreq, rng.normal(loc=0.0, scale=1.0, size=tclen))
    noisewave = rng.normal(loc=0.0, scale=noiseamp, size=tclen)
    theparamsets = [
        [2.0, 0.0, False, 0],
        [2.5, 0.8, True, 0],
        [5.0, 0.5, True, 0],
        [7.5, 0.25, True, 100],
        [10.0, 0.1, True, 0],
    ]
    for paramset in theparamsets:
        echotime = paramset[0]
        echoamp = paramset[1]
        check_autocorrelation = paramset[2]
        osvalidsimcalcstart = paramset[3]
        if debug:
            print(
                "**********Start******************************************************************"
            )
            print(f"{echotime=}, {echoamp=}, {check_autocorrelation=}, {osvalidsimcalcstart=}")
            print(
                "*********************************************************************************"
            )
        theechotc, dummy, dummy, dummy = tide_resample.timeshift(
            basewave, echotime * oversampfreq, padtrs, doplot=displayplots, debug=debug
        )
        resampnonosref_y = basewave + echoamp * theechotc + noisewave
        resampref_y = tide_resample.upsample(resampnonosref_y, fmrifreq, oversampfreq)
        theCorrelator.setreftc(resampnonosref_y)
        referencetc = tide_math.corrnormalize(
            resampref_y[osvalidsimcalcstart:],
            detrendorder=detrendorder,
            windowfunc=windowfunc,
        )

        resampref_y = tide_resample.upsample(resampnonosref_y, fmrifreq, oversampfreq)

        (
            cleaned_resampref_y,
            cleaned_referencetc,
            cleaned_nonosreferencetc,
            despeckle_thresh,
            sidelobeamp,
            sidelobetime,
            lagmod,
            acwidth,
            absmaxsigma,
        ) = tide_cleanregressor.cleanregressor(
            outputname,
            thepass,
            referencetc,
            resampref_y,
            resampnonosref_y,
            fmrifreq,
            oversampfreq,
            osvalidsimcalcstart,
            osvalidsimcalcend,
            lagmininpts,
            lagmaxinpts,
            theFitter,
            theCorrelator,
            lagmin,
            lagmax,
            LGR=None,
            check_autocorrelation=check_autocorrelation,
            fix_autocorrelation=True,
            despeckle_thresh=5.0,
            lthreshval=0.0,
            fixdelay=False,
            detrendorder=detrendorder,
            windowfunc=windowfunc,
            respdelete=False,
            displayplots=displayplots,
            debug=debug,
            rt_floattype=np.float64,
        )
        print(f"\t{len(referencetc)=}")
        print(f"\t{len(resampref_y)=}")
        print(f"\t{len(resampnonosref_y)=}")
        print(f"\t{len(cleaned_resampref_y)=}")
        print(f"\t{len(cleaned_referencetc)=}")
        print(f"\t{len(cleaned_nonosreferencetc)=}")
        print(f"\t{check_autocorrelation=}")
        print(f"\t{despeckle_thresh=}")
        print(f"\t{sidelobeamp=}")
        print(f"\t{sidelobetime=}")
        print(f"\t{lagmod=}")
        print(f"\t{acwidth=}")
        print(f"\t{absmaxsigma=}")
        assert len(referencetc) == len(cleaned_referencetc)
        assert len(resampref_y) == len(cleaned_resampref_y)
        assert len(resampnonosref_y) == len(cleaned_nonosreferencetc)

        if debug:
            print(
                "*********************************************************************************"
            )
            print(f"{echotime=}, {echoamp=}, {check_autocorrelation=}, {osvalidsimcalcstart=}")
            print(
                "**************End****************************************************************"
            )


if __name__ == "__main__":
    mpl.use("TkAgg")
    test_cleanregressor(debug=True, local=True, displayplots=True)
